""" Unit tests for utils module. """

from plone.subrequest.interfaces import ISubRequest
from zope.interface import alsoProvides

import unittest


SITE_LOGO_BASE64 = (
    b"filenameb64:cGl4ZWwucG5n;datab64:iVBORw0KGgoAAAANSUhEUgA"
    b"AAAEAAAABCAIAAACQd1PeAAAADElEQVQI12P4z8AAAAMBAQAY3Y2wAAA"
    b"AAElFTkSuQmCC"
)


class DefaultUtilsTests(unittest.TestCase):
    def test_safe_bytes(self):
        """safe_bytes should always encode unicode to the specified encoding."""
        from plone.base.utils import safe_bytes

        self.assertEqual(safe_bytes("späm"), b"sp\xc3\xa4m")
        self.assertEqual(safe_bytes("späm", "utf-8"), b"sp\xc3\xa4m")
        self.assertEqual(safe_bytes("späm", encoding="latin-1"), b"sp\xe4m")

    def test_get_top_request(self):
        """If in a subrequest, ``get_top_request`` should always return the top
        most request.
        """
        from plone.base.utils import get_top_request

        class MockRequest:
            def __init__(self, parent_request=None):
                self._dict = {}
                if parent_request:
                    self._dict["PARENT_REQUEST"] = parent_request
                    alsoProvides(self, ISubRequest)

            def get(self, key, default=None):
                return self._dict.get(key, default)

        req0 = MockRequest()
        req1 = MockRequest(req0)
        req2 = MockRequest(req1)

        self.assertEqual(get_top_request(req0), req0)
        self.assertEqual(get_top_request(req1), req0)
        self.assertEqual(get_top_request(req2), req0)

    def test_get_top_site_from_url(self):
        """Unit test for ``get_top_site_from_url`` with context and request
        mocks.

        Test content structure:
        /approot/PloneSite/folder/SubSite/folder
        PloneSite and SubSite implement ISite
        """
        from plone.base.utils import get_top_site_from_url
        from urllib.parse import urlparse
        from zope.component.interfaces import ISite

        class MockContext:
            vh_url = "http://nohost"
            vh_root = ""

            def __init__(self, physical_path):
                self.physical_path = physical_path
                if self.physical_path.split("/")[-1] in (
                    "PloneSite",
                    "SubSite",
                ):  # noqa
                    alsoProvides(self, ISite)

            @property
            def id(self):
                return self.physical_path.split("/")[-1]

            def absolute_url(self):
                return (
                    self.vh_url + self.physical_path[len(self.vh_root) :] or "/"
                )  # noqa

            def restrictedTraverse(self, path):
                return MockContext(self.vh_root + path)

            def unrestrictedTraverse(self, path):
                return self.restrictedTraverse(path)

        class MockRequest:
            vh_url = "http://nohost"
            vh_root = ""

            def physicalPathFromURL(self, url):
                # Return the physical path from a URL.
                # The outer right '/' is not part of the path.
                path = self.vh_root + urlparse(url).path.rstrip("/")
                return path.split("/")

        # NO VIRTUAL HOSTING

        req = MockRequest()

        # Case 1:
        ctx = MockContext("/approot/PloneSite")
        self.assertEqual(get_top_site_from_url(ctx, req).id, "PloneSite")

        # Case 2
        ctx = MockContext("/approot/PloneSite/folder")
        self.assertEqual(get_top_site_from_url(ctx, req).id, "PloneSite")

        # Case 3:
        ctx = MockContext("/approot/PloneSite/folder/SubSite/folder")
        self.assertEqual(get_top_site_from_url(ctx, req).id, "PloneSite")

        # Case 4, using unicode paths accidentially:
        ctx = MockContext("/approot/PloneSite/folder/SubSite/folder")
        self.assertEqual(get_top_site_from_url(ctx, req).id, "PloneSite")

        # VIRTUAL HOSTING ON SUBSITE

        req = MockRequest()
        req.vh_root = "/approot/PloneSite/folder/SubSite"

        # Case 4:
        ctx = MockContext("/approot/PloneSite/folder/SubSite")
        ctx.vh_root = "/approot/PloneSite/folder/SubSite"
        self.assertEqual(get_top_site_from_url(ctx, req).id, "SubSite")

        # Case 5:
        ctx = MockContext("/approot/PloneSite/folder/SubSite/folder")
        ctx.vh_root = "/approot/PloneSite/folder/SubSite"
        self.assertEqual(get_top_site_from_url(ctx, req).id, "SubSite")

        # Case 6 (VHM points to child of subsite, this bug existed 4 years):
        req = MockRequest()
        req.vh_root = "/approot/PloneSite/folder/SubSite/en"
        ctx = MockContext("/approot/PloneSite/folder/SubSite/en/archives")
        ctx.vh_root = "/approot/PloneSite/folder/SubSite/en"
        self.assertEqual(get_top_site_from_url(ctx, req).id, "en")

    def test_human_readable_size_int(self):
        from plone.base.utils import human_readable_size

        self.assertEqual(human_readable_size(0), "0 KB")
        self.assertEqual(human_readable_size(1), "1 KB")
        size = 1000
        self.assertEqual(human_readable_size(size), "1 KB")
        size += 24
        self.assertEqual(human_readable_size(size), "1.0 KB")
        size += 512
        self.assertEqual(human_readable_size(size), "1.5 KB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1.5 MB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1.5 GB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1.5 TB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1.5 PB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1536.0 PB")

    def test_human_readable_size_float(self):
        from plone.base.utils import human_readable_size

        self.assertEqual(human_readable_size(0.0), "0 KB")
        self.assertEqual(human_readable_size(1.0), "1 KB")
        size = 1000.0
        self.assertEqual(human_readable_size(size), "1 KB")
        size += 24.0
        self.assertEqual(human_readable_size(size), "1.0 KB")
        size += 512.0
        self.assertEqual(human_readable_size(size), "1.5 KB")
        size *= 1024.0
        self.assertEqual(human_readable_size(size), "1.5 MB")
        size *= 1024.0
        self.assertEqual(human_readable_size(size), "1.5 GB")
        size *= 1024.0
        self.assertEqual(human_readable_size(size), "1.5 TB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1.5 PB")
        size *= 1024
        self.assertEqual(human_readable_size(size), "1536.0 PB")

    def test_human_readable_size_special(self):
        from plone.base.utils import human_readable_size

        self.assertEqual(human_readable_size(None), "0 KB")
        self.assertEqual(human_readable_size(""), "0 KB")
        self.assertEqual(human_readable_size("barney"), "barney")



from .basetool import IPloneBaseTool
from zope.interface import Attribute


class IInterfaceTool(IPloneBaseTool):
    """This tool exposes the interface package for TTW applications,
    by accepting a dotted name of an interface and exporting the
    IInterface API"""

    id = Attribute("id", 'Must be set to "portal_interface"')

    def objectImplements(obj, dotted_name):
        """Asserts if an object implements a given interface"""

    def classImplements(obj, dotted_name):
        """Asserts if an object's class implements a given interface"""

    def namesAndDescriptions(dotted_name, all=0):
        """Returns a list of pairs (name, description) for a given
        interface"""



from .basetool import IPloneBaseTool
from plone import schema
from plone.autoform import directives
from plone.base import PloneMessageFactory as _
from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite
from zope.interface import Attribute
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import Invalid
from zope.interface import invariant
from zope.schema import ValidationError
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import json


def dump_json_to_text(obj):
    """Encode an obj into a text"""
    text = json.dumps(obj, indent=4)
    if not isinstance(text, str):
        text = text.decode("utf8")
    return text


ROBOTS_TXT = """Sitemap: {portal_url}/sitemap.xml.gz

# Define access-restrictions for robots/spiders
# http://www.robotstxt.org/wc/norobots.html



# By default we allow robots to access all areas of our site
# already accessible to anonymous users

User-agent: *
Disallow:



# Add Googlebot-specific syntax extension to exclude forms
# that are repeated for each piece of content in the site
# the wildcard is only supported by Googlebot
# http://www.google.com/support/webmasters/bin/answer.py?answer=40367&ctx=sibling

User-Agent: Googlebot
Disallow: /*?
Disallow: /*atct_album_view$
Disallow: /*folder_factories$
Disallow: /*folder_summary_view$
Disallow: /*login_form$
Disallow: /*mail_password_form$
Disallow: /@@search
Disallow: /*search_rss$
Disallow: /*sendto_form$
Disallow: /*summary_view$
Disallow: /*thumbnail_view$
Disallow: /*view$
"""


def validate_json(value):
    try:
        json.loads(value)
    except ValueError as exc:

        class JSONError(ValidationError):
            __doc__ = _(
                "Must be empty or a valid JSON-formatted "
                "configuration – ${message}.",
                mapping={"message": str(exc)},
            )

        raise JSONError(value)

    return True


class IControlPanel(IPloneBaseTool):
    """Interface for the ControlPanel"""

    def registerConfiglet(
        id,
        name,
        action,
        condition="",
        permission="",  # NOQA
        category="Plone",
        visible=1,
        appId=None,
        imageUrl=None,
        description="",
        REQUEST=None,
    ):
        """Registration of a Configlet"""

    def unregisterConfiglet(id):  # NOQA
        """unregister Configlet"""

    def unregisterApplication(appId):  # NOQA
        """unregister Application with all configlets"""

    def getGroupIds():  # NOQA
        """list of the group ids"""

    def getGroups():  # NOQA
        """list of groups as dicts with id and title"""

    def enumConfiglets(group=None):  # NOQA
        """lists the Configlets of a group, returns them as dicts by
        calling .getAction() on each of them"""


class IEditingSchema(Interface):

    available_editors = schema.List(
        title=_("Available editors"),
        description=_("Available editors in the portal."),
        default=["TinyMCE", "None"],
        value_type=schema.TextLine(),
        missing_value=[],
        required=True,
    )

    default_editor = schema.Choice(
        title=_("Default editor"),
        description=_(
            "Select the default wysiwyg "
            "editor. Users will be able to choose their "
            "own or select to use the site default."
        ),
        default="TinyMCE",
        missing_value=set(),
        vocabulary="plone.app.vocabularies.AvailableEditors",
        required=True,
    )

    ext_editor = schema.Bool(
        title=_("Enable External Editor feature"),
        description=_(
            "Determines if the external editor "
            "feature is enabled. This feature requires a "
            "special client-side application installed. The "
            "users also have to enable this in their "
            "preferences."
        ),
        default=False,
        required=False,
    )

    enable_link_integrity_checks = schema.Bool(
        title=_("Enable link integrity checks"),
        description=_(
            "Determines if the users should get "
            "warnings when they delete or move content that "
            "is linked from inside the site."
        ),
        default=True,
        required=False,
    )

    lock_on_ttw_edit = schema.Bool(
        title=_("Enable locking for through-the-web edits"),
        description=_(
            "Disabling locking here will only "
            "affect users editing content through the "
            "Plone web UI.  Content edited via WebDAV "
            "clients will still be subject to locking."
        ),
        default=True,
        required=False,
    )

    subjects_of_navigation_root = schema.Bool(
        title=_("Limit tags/keywords to the current navigation root"),
        description=_(
            "Limit tags aka keywords vocabulary used for Tags field and "
            "in searches to the terms used inside the subtree of the current "
            "navigation root. This can be used together with Plone's "
            "multilingual extension plone.app.multilingual to only offer "
            "keywords of the current selected language. Other addons may "
            "utilize this feature for its specific purposes."
        ),
        default=False,
        required=False,
    )


class ITagAttrPair(Interface):
    tags = schema.TextLine(title="tags")
    attributes = schema.TextLine(title="attributes")


@implementer(ITagAttrPair)
class TagAttrPair:
    def __init__(self, tags="", attributes=""):
        self.tags = tags
        self.attributes = attributes


class IFilterSchema(Interface):
    """Combined schema for the adapter lookup."""

    # class IFilterTagsSchema(Interface):

    disable_filtering = schema.Bool(
        title=_("Disable HTML filtering"),
        description=_(
            "Warning: disabling this can be dangerous. "
            "Only disable if you know what you are doing."
        ),
        default=False,
        required=False,
    )

    nasty_tags = schema.List(
        title=_("Nasty tags"),
        description=_(
            "These tags and their content are completely blocked "
            "when a page is saved or rendered. They are only deleted"
            " if they are not marked as valid_tags"
        ),
        default=["style", "object", "embed", "applet", "script", "meta"],
        value_type=schema.TextLine(),
        missing_value=[],
        required=False,
    )

    valid_tags = schema.List(
        title=_("Valid tags"),
        description=_("A list of valid tags which will be not filtered out."),
        default=[
            "a",
            "abbr",
            "acronym",
            "address",
            "article",
            "aside",
            "audio",
            "b",
            "bdo",
            "big",
            "blockquote",
            "body",
            "br",
            "canvas",
            "caption",
            "cite",
            "code",
            "col",
            "colgroup",
            "command",
            "datalist",
            "dd",
            "del",
            "details",
            "dfn",
            "dialog",
            "div",
            "dl",
            "dt",
            "em",
            "figure",
            "figcaption",
            "footer",
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
            "head",
            "header",
            "hgroup",
            "hr",
            "html",
            "i",
            "iframe",
            "img",
            "ins",
            "kbd",
            "keygen",
            "li",
            "map",
            "mark",
            "meter",
            "nav",
            "ol",
            "output",
            "p",
            "pre",
            "picture",
            "progress",
            "q",
            "rp",
            "rt",
            "ruby",
            "samp",
            "section",
            "small",
            "source",
            "span",
            "strong",
            "sub",
            "sup",
            "table",
            "tbody",
            "td",
            "tfoot",
            "th",
            "thead",
            "time",
            "title",
            "tr",
            "tt",
            "u",
            "ul",
            "var",
            "video",
        ],
        value_type=schema.TextLine(),
        missing_value=[],
        required=False,
    )

    custom_attributes = schema.List(
        title=_("Custom attributes"),
        description=_("These attributes are additionally allowed."),
        default=["style", "controls", "poster", "autoplay", "loading", "srcset", "sizes"],
        value_type=schema.TextLine(),
        missing_value=[],
        required=False,
    )


class ITinyMCELayoutSchema(Interface):
    """This interface defines the layout properties."""

    resizing = schema.Bool(
        title=_("Enable resizing the editor window."),
        description=_(
            "This option gives you the ability to enable/disable "
            "resizing the editor window. "
        ),
        default=True,
        required=False,
    )

    autoresize = schema.Bool(
        title=_("Enable auto resizing of the editor window."),
        description=_(
            "This option gives you the ability to enable/disable "
            "auto resizing the editor window depending "
            "on the content."
        ),
        default=False,
        required=False,
    )

    inline = schema.Bool(
        title=_("Run TinyMCE in inline mode."),
        description=_(
            "The inline editing mode is useful when creating user "
            "experiences where you want the editing view of the page "
            "to be merged with the reading view of the page."
        ),
        default=False,
        required=False,
    )

    # TODO: add validation to assert % and px in the value
    editor_width = schema.TextLine(
        title=_("Editor width"),
        description=_(
            "This option gives you the ability to specify the "
            "width of the editor (like 100% or 400px)."
        ),
        default=None,
        required=False,
    )

    # TODO: add validation to assert % and px in the value
    editor_height = schema.TextLine(
        title=_("Editor height"),
        description=_(
            "This option gives you the ability to specify the "
            "height of the editor in pixels. "
            "If auto resize is enabled this value is used "
            "as minimum height."
        ),
        default=None,
        required=False,
    )

    content_css = schema.List(
        title=_("Choose the CSS used in WYSIWYG Editor Area"),
        description=_(
            "This option enables you to specify CSS files "
            "that will be used within the the editable area of the editor "
            "(e.g. ++plone++mystyles/tinymce.css). "
            "These styles will also appear in the Format list of TinyMCE."
            "In addition to what is listed here, "
            "the CSS file defined under 'tinymce-content-css'"
            " in a theme manifest.cfg is also added."
        ),
        value_type=schema.TextLine(),
        missing_value=[],
        default=[],
        required=False,
    )

    header_styles = schema.List(
        title=_("Header styles"),
        description=_("Name|tag"),
        value_type=schema.TextLine(),
        missing_value=[],
        default=[
            "Header 1|h1",
            "Header 2|h2",
            "Header 3|h3",
            "Header 4|h4",
            "Header 5|h5",
            "Header 6|h6",
        ],
    )

    inline_styles = schema.List(
        title=_("Inline styles"),
        description=_("Name|format|icon"),
        value_type=schema.TextLine(),
        missing_value=[],
        default=[
            "Bold|bold|bold",
            "Italic|italic|italic",
            "Underline|underline|underline",
            "Strikethrough|strikethrough|strikethrough",
            "Superscript|superscript|superscript",
            "Subscript|subscript|subscript",
            "Code|code|code",
            "Text in 2 columns|textcolumns2",
            "Text in 3 columns|textcolumns3",
        ],
    )

    block_styles = schema.List(
        title=_("Block styles"),
        description=_("Name|format"),
        value_type=schema.TextLine(),
        missing_value=[],
        default=["Paragraph|p", "Blockquote|blockquote", "Div|div", "Pre|pre"],
    )

    alignment_styles = schema.List(
        title=_("Alignment styles"),
        description=_("Name|format|icon"),
        value_type=schema.TextLine(),
        missing_value=[],
        default=[
            "Left|alignleft|align-left",
            "Center|aligncenter|align-center",
            "Right|alignright|align-right",
            "Justify|alignjustify|align-justify",
        ],
    )

    table_styles = schema.List(
        title=_("Table styles"),
        description=_("Name|class"),
        value_type=schema.TextLine(),
        missing_value=[],
        default=["Listing|listing", "Listing compact|listing-compact", "Subdued grid|plain", "Invisible Grid|invisible-grid"],
    )

    formats = schema.Text(
        title=_("Formats"),
        description=_(
            "Enter a JSON-formatted style format configuration. "
            "A format is for example the style that get applied when "
            "you press the bold button inside the editor. "
            "See https://www.tinymce.com/docs/configure/content-formatting/#formats"
        ),  # NOQA: E501
        constraint=validate_json,
        default=dump_json_to_text(
            {
                "discreet": {"inline": "span", "classes": "discreet"},
                "clearfix": {"block": "div", "classes": "clearfix"},
                "alignleft": {
                    "selector": "p,h1,h2,h3,h4,h5,h6,td,th,div,ul,ol,li,table",
                    "classes": "text-start",
                },
                "aligncenter": {
                    "selector": "p,h1,h2,h3,h4,h5,h6,td,th,div,ul,ol,li,table",
                    "classes": "text-center",
                },
                "alignright": {
                    "selector": "p,h1,h2,h3,h4,h5,h6,td,th,div,ul,ol,li,table",
                    "classes": "text-end",
                },
                "alignjustify": {
                    "selector": "p,h1,h2,h3,h4,h5,h6,td,th,div,ul,ol,li,table",
                    "classes": "text-justify",
                },
                "textcolumns2": {"selector": "p", "classes": "text-columns-2"},
                "textcolumns3": {"selector": "p", "classes": "text-columns-3"},
            }
        ),
        required=True,
    )


class ITinyMCEPluginSchema(Interface):
    """This interface defines the toolbar properties."""

    plugins = schema.List(
        title=_("label_tinymce_plugins", default="Editor plugins"),
        description=_(
            "help_tinymce_plugins", default=("Select plugins to include with tinymce")
        ),
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary(
                [
                    SimpleTerm("advlist", "advlist", "advlist"),
                    SimpleTerm("anchor", "anchor", "anchor"),
                    SimpleTerm("autolink", "autolink", "autolink"),
                    SimpleTerm("autosave", "autosave", "autosave"),
                    SimpleTerm("charmap", "charmap", "charmap"),
                    SimpleTerm("code", "code", "code"),
                    SimpleTerm("colorpicker", "colorpicker", "colorpicker"),
                    SimpleTerm("contextmenu", "contextmenu", "contextmenu"),
                    SimpleTerm("directionality", "directionality", "directionality"),
                    SimpleTerm("emoticons", "emoticons", "emoticons"),
                    SimpleTerm("fullpage", "fullpage", "fullpage"),
                    SimpleTerm("fullscreen", "fullscreen", "fullscreen"),
                    SimpleTerm("hr", "hr", "hr"),
                    SimpleTerm("insertdatetime", "insertdatetime", "insertdatetime"),
                    SimpleTerm("layer", "layer", "layer"),
                    SimpleTerm("lists", "lists", "lists"),
                    SimpleTerm("media", "media", "media"),
                    SimpleTerm("nonbreaking", "nonbreaking", "nonbreaking"),
                    SimpleTerm("noneditable", "noneditable", "noneditable"),
                    SimpleTerm("pagebreak", "pagebreak", "pagebreak"),
                    SimpleTerm("paste", "paste", "paste"),
                    SimpleTerm("preview", "preview", "preview"),
                    SimpleTerm("print", "print", "print"),
                    # XXX disable save button since it is not implemeneted
                    # SimpleTerm('save', 'save', u'save'),
                    SimpleTerm("searchreplace", "searchreplace", "searchreplace"),
                    SimpleTerm("tabfocus", "tabfocus", "tabfocus"),
                    SimpleTerm("table", "table", "table"),
                    SimpleTerm("textcolor", "textcolor", "textcolor"),
                    SimpleTerm("textpattern", "textpattern", "textpattern"),
                    SimpleTerm("template", "template", "template"),
                    SimpleTerm("visualblocks", "visualblocks", "visualblocks"),
                    SimpleTerm("visualchars", "visualchars", "visualchars"),
                    SimpleTerm("wordcount", "wordcount", "wordcount"),
                ]
            )
        ),
        default=[
            "fullscreen",
            "hr",
            "lists",
            "media",
            "nonbreaking",
            "noneditable",
            "pagebreak",
            "paste",
            "preview",
            "print",
            "searchreplace",
            "tabfocus",
            "table",
            "visualchars",
            "wordcount",
            "code",
        ],
        missing_value=[],
        required=False,
    )

    menubar = schema.List(
        title=_("label_tinymce_menubar", default="Menubar"),
        description=_(
            "help_tinymce_menubar",
            default=("Enter what items you would like in the menu bar."),
        ),
        required=True,
        value_type=schema.TextLine(),
        missing_value=[],
        default=["edit", "table", "format", "tools", "view", "insert"],
    )

    menu = schema.Text(
        title=_("label_tinymce_menu", "Menu"),
        description=_(
            "hint_tinymce_menu", default="JSON formatted Menu configuration."
        ),
        constraint=validate_json,
        default=dump_json_to_text(
            {
                "edit": {
                    "title": "Edit",
                    "items": "undo redo | cut copy paste pastetext | "
                    "searchreplace textpattern selectall | textcolor",
                },
                "insert": {"title": "Insert", "items": "link media | template hr"},
                "view": {
                    "title": "View",
                    "items": "visualaid visualchars visualblocks preview "
                    "fullpage fullscreen code",
                },
                "format": {
                    "title": "Format",
                    "items": "bold italic underline strikethrough "
                    "superscript subscript | formats | removeformat",
                },
                "table": {
                    "title": "Table",
                    "items": "inserttable tableprops deletetable | cell row column",
                },
                "tools": {
                    "title": "Tools",
                    "items": "spellchecker charmap emoticons insertdatetime layer",
                },
            }
        ),
    )

    templates = schema.Text(
        title=_("label_tinymce_templates", default="Templates"),
        description=_(
            "help_tinymce_templates",
            default=(
                "Enter the list of templates in json format "
                "https://www.tinymce.com/docs/plugins/template/"
            ),
        ),
        required=False,
        constraint=validate_json,
        default=dump_json_to_text({}),
    )

    toolbar = schema.Text(
        title=_("label_tinymce_toolbar", default="Toolbar"),
        description=_(
            "help_tinymce_toolbar",
            default=("Enter how you would like the toolbar items to list."),
        ),
        required=True,
        default="ltr rtl | undo redo | styleselect | bold italic | "
        "alignleft aligncenter alignright alignjustify | "
        "bullist numlist outdent indent | "
        "inserttable | unlink plonelink ploneimage",
    )

    custom_plugins = schema.List(
        title=_("Custom plugins"),
        description=_(
            "Enter a list of custom plugins which will be loaded "
            "in the editor. Format is "
            "pluginname|location, one per line."
        ),
        required=False,
        value_type=schema.TextLine(),
        missing_value=[],
        default=[],
    )

    custom_buttons = schema.List(
        title=_("Custom buttons"),
        description=_("Enter a list of custom buttons which will be added to toolbar"),
        required=False,
        value_type=schema.TextLine(),
        missing_value=[],
        default=[],
    )


ITinyMCELibrariesSchema = ITinyMCEPluginSchema  # bw compat


class ITinyMCESpellCheckerSchema(Interface):
    """This interface defines the libraries properties."""

    libraries_spellchecker_choice = schema.Choice(
        title=_("Spellchecker plugin to use"),
        description=_("This option allows you to choose the spellchecker for TinyMCE."),
        missing_value=set(),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("browser", "browser", _("Default browser spellchecker")),
                SimpleTerm("AtD", "AtD", _("After the deadline (FLOSS)")),
            ]
        ),
        default="browser",
        required=False,
    )

    libraries_atd_ignore_strings = schema.List(
        title=_("AtD ignore strings"),
        description=_(
            "label_atd_ignore_strings",
            default='A list of strings which the "After the Deadline" '
            "spellchecker should ignore. "
            "Note: This option is only applicable when the "
            "appropriate spellchecker has been chosen above.",
        ),
        default=["Zope", "Plone", "TinyMCE"],
        value_type=schema.TextLine(),
        missing_value=[],
        required=False,
    )

    libraries_atd_show_types = schema.List(
        title=_("AtD error types to show"),
        description=_(
            "help_atderrortypes_to_show",
            default="A list of error types which the "
            '"After the Deadline" spellchecker should check for. '
            "By default, all the available error type will be "
            "listed here.",
        ),
        value_type=schema.TextLine(),
        default=[
            "Bias Language",
            "Cliches",
            "Complex Expression",
            "Diacritical Marks",
            "Double Negatives",
            "Hidden Verbs",
            "Jargon Language",
            "Passive voice",
            "Phrases to Avoid",
            "Redundant Expression",
        ],
        missing_value=[],
        required=False,
    )

    libraries_atd_service_url = schema.TextLine(
        title=_("AtD service URL"),
        description=_(
            "help_atd_service_url",
            default='The URL of the "After the Deadline" grammar and spell '
            "checking server. "
            "The default value is the public server, "
            "but ideally you should download and install your own "
            "and specify its address here.",
        ),
        required=True,
        default="service.afterthedeadline.com",
    )


class ITinyMCEResourceTypesSchema(Interface):
    """This interface defines the resource types properties."""

    # XXX Not implemented in new tinymce version. Need to decide about this
    # rooted = schema.Bool(
    #    title=_(u"Rooted to current object"),
    #    description=_(u"When enabled the user will be rooted to the current "
    #                  "object and can't add links and images from other "
    #                  "parts of the site."),
    #    default=False,
    #    required=False)

    contains_objects = schema.List(
        title=_("Contains objects"),
        description=_(
            "Enter a list of content types which can contain other "
            "objects. Format is one contenttype per line."
        ),
        value_type=schema.TextLine(),
        default=["Folder", "Large Plone Folder", "Plone Site"],
        missing_value=[],
        required=False,
    )

    image_objects = schema.List(
        title=_("Image objects"),
        description=_(
            "Enter a list of content types which can be used as "
            "images. Format is one contenttype per line."
        ),
        default=["Image"],
        value_type=schema.TextLine(),
        missing_value=[],
        required=False,
    )

    entity_encoding = schema.Choice(
        title=_("Entity encoding"),
        description=_(
            "This option controls how entities/characters get processed. "
            "Named: Characters will be converted into named entities "
            "based on the entities option. "
            "Numeric: Characters will be converted into numeric entities. "
            "Raw: All characters will be stored in non-entity form "
            "except these XML default entities: amp lt gt quot"
        ),
        # missing_value=set(),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("named", "named", _("Named")),
                SimpleTerm("numeric", "numeric", _("Numeric")),
                SimpleTerm("raw", "raw", _("Raw")),
            ]
        ),
        default="raw",
        required=False,
    )


class ITinyMCEAdvancedSchema(Interface):
    """This interface defines the resource types properties."""

    other_settings = schema.Text(
        title=_("label_tinymce_other_settings", "Other settings"),
        description=_(
            "hint_tinymce_other_settings",
            default="Other TinyMCE configuration formatted as JSON.",
        ),
        required=False,
        constraint=validate_json,
        default=dump_json_to_text({}),
    )


class ITinyMCESchema(
    ITinyMCELayoutSchema,
    ITinyMCEPluginSchema,
    ITinyMCESpellCheckerSchema,
    ITinyMCEResourceTypesSchema,
    ITinyMCEAdvancedSchema,
):
    """TinyMCE Schema"""


class IMaintenanceSchema(Interface):

    days = schema.Int(
        title=_("Days of object history to keep after packing"),
        description=_(
            "You should pack your database regularly. This number "
            "indicates how many days of undo history you want to "
            "keep. It is unrelated to versioning, so even if you "
            "pack the database, the history of the content changes "
            "will be kept. Recommended value is 7 days."
        ),
        default=7,
        min=0,
        required=True,
    )


class INavigationSchema(Interface):

    navigation_depth = schema.Int(
        title=_("Navigation depth"),
        description=_("Number of folder levels to show in the navigation."),
        default=3,
        required=True,
    )

    generate_tabs = schema.Bool(
        title=_("Automatically generate tabs"),
        description=_(
            "By default, all items created at the root level will "
            "appear as tabs. You can turn this off if you prefer manually "
            "constructing this part of the navigation."
        ),
        default=True,
        required=False,
    )

    nonfolderish_tabs = schema.Bool(
        title=_("Generate tabs for items other than folders."),
        description=_(
            "By default, any content item in the root of the portal will "
            "appear as a tab. If you turn this option off, only folders "
            "will be shown. This only has an effect if 'automatically "
            "generate tabs' is enabled."
        ),
        default=True,
        required=False,
    )

    sort_tabs_on = schema.Choice(
        title=_("Sort tabs on"),
        description=_("Index used to sort the tabs"),
        required=True,
        default="getObjPositionInParent",
        vocabulary=SimpleVocabulary(
            [
                # there is no vocabulary of sortable indexes by now, so hard code
                # some options here
                SimpleTerm(
                    "getObjPositionInParent",
                    "getObjPositionInParent",
                    _("Position in Parent"),
                ),
                SimpleTerm("sortable_title", "sortable_title", _("Title")),
                SimpleTerm("getId", "getId", _("Short Name (ID)")),
            ]
        ),
    )
    sort_tabs_reversed = schema.Bool(
        title=_("Reversed sort order for tabs."),
        description=_("Sort tabs in descending."),
        default=False,
        required=False,
    )

    displayed_types = schema.Tuple(
        title=_("Displayed content types"),
        description=_(
            "The content types that should be shown in the navigation and site map."
        ),
        required=False,
        default=("Link", "News Item", "Folder", "Document", "Event", "Collection"),
        missing_value=(),
        value_type=schema.Choice(
            source="plone.app.vocabularies.ReallyUserFriendlyTypes"
        ),
    )

    filter_on_workflow = schema.Bool(
        title=_("Filter on workflow state"),
        description=_(
            "The workflow states that should be shown in the navigation "
            "and the site map."
        ),
        default=False,
        required=False,
    )

    workflow_states_to_show = schema.Tuple(
        required=False,
        default=(),
        missing_value=(),
        value_type=schema.Choice(source="plone.app.vocabularies.WorkflowStates"),
    )

    show_excluded_items = schema.Bool(
        title=_(
            "Show items normally excluded from navigation if viewing their children."
        ),
        description=_(
            "If an item has been excluded from navigation should it be "
            "shown in navigation when viewing content contained within it "
            "or within a subfolder."
        ),
        default=False,
        required=False,
    )

    root = schema.TextLine(
        title=_("Root"),
        description=_(
            "Path to be used as navigation root, relative to Plone site root."
            "Starts with '/'"
        ),
        default="/",
        required=True,
    )

    sitemap_depth = schema.Int(
        title=_("Sitemap depth"),
        description=_("Number of folder levels to show in the site map."),
        default=3,
        required=True,
    )

    parent_types_not_to_query = schema.List(
        title=_("Hide children of these types"),
        description=_("Hide content inside the following types in Navigation."),
        default=["TempFolder"],
        value_type=schema.TextLine(),
        missing_value=(),
        required=False,
    )


class ISearchSchema(Interface):

    enable_livesearch = schema.Bool(
        title=_("Enable LiveSearch"),
        description=_(
            "Enables the LiveSearch feature, which shows live "
            "results if the browser supports JavaScript."
        ),
        default=True,
        required=False,
    )

    types_not_searched = schema.Tuple(
        title=_("Select content types which should be excluded from search results"),
        description=_(
            "Select the content types that should be excluded from search results. "
            "Note that if new content types are installed, they "
            "will be enabled by default unless explicitly turned "
            "off here or by the relevant installer."
        ),
        required=False,
        default=(
            "Discussion Item",
            "Plone Site",
            "TempFolder",
        ),
        missing_value=(),
        value_type=schema.Choice(source="plone.app.vocabularies.PortalTypes"),
    )

    search_results_description_length = schema.Int(
        title=_(
            "Crop the item description in search result listings "
            "after a number of characters."
        ),
        required=False,
        default=160,
    )

    sort_on = schema.Choice(
        title=_("label_sort_on", default="Sort on"),
        description=_("Sort the default search on this index"),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("relevance", "relevance", _("relevance")),
                SimpleTerm("Date", "Date", _("date (newest first)")),
                SimpleTerm("sortable_title", "sortable_title", _("alphabetically")),
            ]
        ),
        default="relevance",
        required=True,
    )

    search_show_images = schema.Bool(
        title=_("Show images in results"),
        default=True,
        required=False,
    )

    search_image_scale = schema.Choice(
        title=_("Image scale for results"),
        default="mini",
        vocabulary="plone.app.vocabularies.ImagesScales",
        required=False,
    )


class ISecuritySchema(Interface):

    enable_self_reg = schema.Bool(
        title=_("Enable self-registration"),
        description=_(
            "Allows users to register themselves on the site. If "
            "not selected, only site managers can add new users."
        ),
        default=False,
        required=False,
    )

    enable_user_pwd_choice = schema.Bool(
        title=_("Let users select their own passwords"),
        description=_(
            "If not selected, a URL will be generated and "
            "e-mailed. Users are instructed to follow the link to "
            "reach a page where they can change their password and "
            "complete the registration process; this also verifies "
            "that they have entered a valid email address."
        ),
        default=False,
        required=False,
    )

    enable_user_folders = schema.Bool(
        title=_("Enable User Folders"),
        description=_(
            "If selected, home folders where users can create "
            "content will be created when they log in."
        ),
        default=False,
        required=False,
    )

    allow_anon_views_about = schema.Bool(
        title=_("Allow anyone to view 'about' information"),
        description=_(
            "If not selected only logged-in users will be able to "
            "view information about who created an item and when it "
            "was modified."
        ),
        default=False,
        required=False,
    )

    use_email_as_login = schema.Bool(
        title=_("Use email address as login name"),
        description=_(
            "Allows users to login with their email address instead "
            "of specifying a separate login name. This also updates "
            "the login name of existing users, which may take a "
            "while on large sites. The login name is saved as "
            "lower case, but to be userfriendly it does not matter "
            "which case you use to login. When duplicates are found, "
            "saving this form will fail. You can use the "
            "@@migrate-to-emaillogin page to show the duplicates."
        ),
        default=False,
        required=False,
    )

    use_uuid_as_userid = schema.Bool(
        title=_("Use UUID user ids"),
        description=_(
            "Use automatically generated UUIDs as user id for new users. "
            "When not turned on, the default is to use the same as the "
            "login name, or when using the email address as login name we "
            "generate a user id based on the fullname."
        ),
        default=False,
        required=False,
    )

    autologin_after_password_reset = schema.Bool(
        title=_("Login user after password reset"),
        description=_(
            "After successful password reset the user will be logged "
            "in automatically."
        ),
        default=True,
        required=False,
    )


class ISiteSchema(Interface):

    site_title = schema.TextLine(
        title=_("Site title"),
        description=_(
            "This shows up in the title bar of browsers and in syndication feeds."
        ),
        default="Plone site",
    )

    site_logo = schema.Bytes(
        title=_("Site Logo"),
        description=_("This shows a custom logo on your site."),
        required=False,
    )

    site_favicon_mimetype = schema.TextLine(
        title=_("MIME type of the site favicon"),
        description=_(
            "MIME type of the favicon (automatically set when a new favicon is uploaded)"
        ),
        required=False,
        default="image/vnd.microsoft.icon",
    )

    site_favicon = schema.Bytes(
        title=_("Site Favicon"),
        description=_("This shows a custom favicon on your site."),
        required=False,
    )

    exposeDCMetaTags = schema.Bool(
        title=_("Expose Dublin Core metadata"),
        description=_("Exposes the Dublin Core properties as metatags."),
        default=False,
        required=False,
    )

    enable_sitemap = schema.Bool(
        title=_("Expose sitemap.xml.gz"),
        description=_(
            "Exposes your content as a file "
            "according to the "
            "<a href='http://sitemaps.org'>sitemaps.org</a> "
            "standard. You "
            "can submit this to compliant search engines "
            "like Google, Yahoo and Microsoft. It allows "
            "these search engines to more intelligently "
            "crawl your site."
        ),
        default=False,
        required=False,
    )

    webstats_js = schema.SourceText(
        title=_("JavaScript for web statistics support"),
        description=_(
            "For enabling web statistics support "
            "from external providers (e.g. Google "
            "Analytics). Paste the provided code snippet here. "
            "It will be rendered as "
            "entered near the end of the page."
        ),
        default="",
        required=False,
    )

    display_publication_date_in_byline = schema.Bool(
        title=_("Display publication date"),
        description=_("Show in the byline the date a content item was published."),
        default=False,
        required=False,
    )

    icon_visibility = schema.Choice(
        title=_("Icon visibility"),
        description=_("Show icons in listings"),
        default="enabled",
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("false", "false", _("Never")),
                SimpleTerm("enabled", "enabled", _("Always")),
                SimpleTerm(
                    "authenticated", "authenticated", _("For authenticated users only")
                ),
            ]
        ),
        required=True,
    )

    thumb_visibility = schema.Choice(
        title=_("Thumb visibility"),
        description=_("Show thumbnail images in listings"),
        default="enabled",
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("false", "false", _("Never")),
                SimpleTerm("enabled", "enabled", _("Always")),
                SimpleTerm(
                    "authenticated", "authenticated", _("For authenticated users only")
                ),
            ]
        ),
        required=True,
    )

    no_thumbs_portlet = schema.Bool(
        title=_("No Thumbs in portlets"),
        description=_(
            "Suppress thumbs in all portlets;"
            " this default can be overridden individually "
            "in selected portlets"
        ),
        default=False,
        required=False,
    )

    no_thumbs_lists = schema.Bool(
        title=_("No thumbs in list views"),
        description=_(
            "Suppress thumbs in all list views; "
            "this default can be overriden individually"
        ),
        default=False,
        required=False,
    )

    no_thumbs_summary = schema.Bool(
        title=_("No thumbs in summary views"),
        description=_(
            "Suppress thumbs in all summary views; "
            "this default can be overriden individually"
        ),
        default=False,
        required=False,
    )

    no_thumbs_tables = schema.Bool(
        title=_("No thumbs in table views"),
        description=_(
            "Suppress thumbs in all tableviews and in folder contents view; "
            "this default can be overriden individually"
        ),
        default=False,
        required=False,
    )

    thumb_scale_portlet = schema.Choice(
        title=_("Thumb scale for portlets"),
        description=_("This default can be overriden individually."),
        default="icon",
        vocabulary="plone.app.vocabularies.ImagesScales",
        required=True,
    )

    thumb_scale_listing = schema.Choice(
        title=_("Thumb scale for listings"),
        description=_(
            "E.g. standard view; This default can be overriden individually."
        ),
        default="thumb",
        vocabulary="plone.app.vocabularies.ImagesScales",
        required=True,
    )

    thumb_scale_table = schema.Choice(
        title=_("Thumb scale for tables"),
        description=_(
            "E.g., tabular view, folder content listing;"
            " This default can be overriden individually."
        ),
        default="tile",
        vocabulary="plone.app.vocabularies.ImagesScales",
        required=True,
    )

    thumb_scale_summary = schema.Choice(
        title=_("Thumb scale for summary view"),
        description=_("This default can be overriden individually."),
        default="mini",
        vocabulary="plone.app.vocabularies.ImagesScales",
        required=True,
    )

    toolbar_position = schema.Choice(
        title=_("Toolbar position"),
        description=_(
            "It can be on the side (vertical mode) or on the top (horizontal mode)"
        ),
        default="side",
        vocabulary=SimpleVocabulary(
            [SimpleTerm("side", "side", _("Side")), SimpleTerm("top", "top", _("Top"))]
        ),
        required=True,
    )

    toolbar_logo = schema.TextLine(
        title=_("Relative URL for the toolbar logo"),
        description=_(
            "This must be a URL relative to the site root. "
            "By default it is /++plone++static/plone-toolbarlogo.svg"
        ),
        default="/++plone++static/plone-toolbarlogo.svg",
        required=False,
    )

    robots_txt = schema.SourceText(
        title=_("robots.txt"),
        description=_(
            "help_robots_txt",
            default="robots.txt is read by search engines to "
            "determine how to index your site. "
            "For details see <a href='http://www.robotstxt.org'>"
            "http://www.robotstxt.org</a>. "
            "Use '{portal_url}' for the site URL.",
        ),
        default=ROBOTS_TXT,
        required=False,
    )

    default_page = schema.List(
        title=_("Default page IDs"),
        description=_(
            "Select which IDs (short names) can act as fallback "
            "default pages for a container."
        ),
        required=False,
        default=["index_html", "index.html", "index.htm", "FrontPage"],
        missing_value=[],
        value_type=schema.TextLine(),
    )

    roles_allowed_to_add_keywords = schema.List(
        title=_("Roles that can add keywords"),
        description=_(
            "help_allow_roles_to_add_keywords",
            default="Only the following roles can add new keywords ",
        ),
        required=False,
        default=[
            "Manager",
            "Site Administrator",
            "Reviewer",
        ],
        missing_value=[],
        value_type=schema.Choice(vocabulary="plone.app.vocabularies.Roles"),
    )


class IDateAndTimeSchema(Interface):
    """Controlpanel settings for date and time related settings."""

    portal_timezone = schema.Choice(
        title=_("Portal default timezone"),
        description=_(
            "help_portal_timezone",
            default="The timezone setting of the portal. Users can set "
            "their own timezone, if available timezones are "
            "defined.",
        ),
        required=True,
        default=None,
        vocabulary="plone.app.vocabularies.CommonTimezones",
    )

    available_timezones = schema.List(
        title=_("Available timezones"),
        description=_(
            "help_available_timezones",
            default="The timezones, which should be available for the "
            "portal. Can be set for users and events",
        ),
        required=False,
        default=[],
        missing_value=[],
        value_type=schema.Choice(vocabulary="plone.app.vocabularies.Timezones"),
    )

    first_weekday = schema.Choice(
        title=_("label_first_weekday", default="First weekday"),
        description=_("help_first_weekday", default="First day in the week."),
        required=True,
        default=None,
        vocabulary="plone.app.vocabularies.Weekdays",
    )


class ITypesSchema(Interface):
    """Controlpanel settings for the types settings."""

    types_use_view_action_in_listings = schema.List(
        title=_("Types which use the view action in listing views."),
        description=_(
            "help_types_use_view_action_in_listings",
            default="When clicking items in listing views, these "
            "types will use the 'view' action instead of using "
            "their default view.",
        ),
        required=False,
        default=["Image", "File"],
        missing_value=[],
        value_type=schema.TextLine(),
    )

    redirect_links = schema.Bool(
        title=_("Redirect links"),
        description=_(
            "help_redirect_links",
            default="When clicking on a Link type, should the user be "
            "taken to the default view or be redirected to the "
            "Link's URL?",
        ),
        required=False,
        default=True,
    )

    default_page_types = schema.List(
        title=_("Types that can be set as a default page"),
        description=_(
            "The content types that should be available for selection "
            "when setting a default page."
        ),
        required=False,
        missing_value=[],
        default=[
            "Document",
            "Event",
            "News Item",
            "Collection",
        ],
        value_type=schema.TextLine(),
    )


class IMailSchema(Interface):

    smtp_host = schema.TextLine(
        title=_("label_smtp_server", default="SMTP server"),
        description=_(
            "help_smtp_server",
            default="The address of your local "
            "SMTP (outgoing e-mail) server. Usually "
            "'localhost', unless you use an "
            "external server to send e-mail.",
        ),
        default="localhost",
        required=True,
    )

    smtp_port = schema.Int(
        title=_("label_smtp_port", default="SMTP port"),
        description=_(
            "help_smtp_port",
            default="The port of your local SMTP "
            "(outgoing e-mail) server. Usually '25'.",
        ),
        default=25,
        required=True,
    )

    smtp_userid = schema.TextLine(
        title=_("label_smtp_userid", default="ESMTP username"),
        description=_(
            "help_smtp_userid",
            default="Username for authentication "
            "to your e-mail server. Not required "
            "unless you are using ESMTP.",
        ),
        default=None,
        required=False,
    )

    smtp_pass = schema.Password(
        title=_("label_smtp_pass", default="ESMTP password"),
        description=_(
            "help_smtp_pass", default="The password for the ESMTP user account."
        ),
        default=None,
        required=False,
    )

    email_from_name = schema.TextLine(
        title=_("Site 'From' name"),
        description=_("Plone generates e-mail using this name as the e-mail sender."),
        default=None,
        required=True,
    )

    email_from_address = schema.ASCIILine(
        title=_("Site 'From' address"),
        description=_(
            "Plone generates e-mail using "
            "this address as the e-mail "
            "return address. It is also "
            "used as the destination "
            "address for the site-wide "
            "contact form and the 'Send test "
            "e-mail' feature."
        ),
        default=None,
        required=True,
    )

    email_charset = schema.ASCIILine(
        title=_("E-mail characterset"),
        description=_("Characterset to use when sending e-mails."),
        default="utf-8",
        required=True,
    )


class IMarkupSchema(Interface):

    default_type = schema.Choice(
        title=_("Default format"),
        description=_(
            "Select the default format of textfields for newly "
            "created content objects."
        ),
        default="text/html",
        vocabulary="plone.app.vocabularies.AllowableContentTypes",
        required=True,
    )

    allowed_types = schema.Tuple(
        title=_("Alternative formats"),
        description=_(
            "Select which formats are available for users as "
            "alternative to the default format. Note that if new "
            "formats are installed, they will be enabled for text "
            "fields by default unless explicitly turned off here "
            "or by the relevant installer."
        ),
        required=True,
        default=("text/html", "text/x-web-textile"),
        missing_value=(),
        value_type=schema.Choice(
            vocabulary="plone.app.vocabularies.AllowableContentTypes"
        ),
    )

    markdown_extensions = schema.List(
        default=[
            "markdown.extensions.fenced_code",
            "markdown.extensions.nl2br",
        ],
        description=_(
            "Look for available extensions at "
            "https://python-markdown.github.io/extensions/ or write your own."
        ),
        missing_value=(),
        required=False,
        title=_("Enabled markdown extensions"),
        value_type=schema.TextLine(),
    )


class IUserGroupsSettingsSchema(Interface):

    many_groups = schema.Bool(
        title=_("Many groups?"),
        description=_(
            "Determines if your Plone is optimized "
            "for small or large sites. In environments with a "
            "lot of groups it can be very slow or impossible "
            "to build a list all groups. This option tunes the "
            "user interface and behaviour of Plone for this "
            "case by allowing you to search for groups instead "
            "of listing all of them."
        ),
        default=False,
        required=False,
    )

    many_users = schema.Bool(
        title=_("Many users?"),
        description=_(
            "Determines if your Plone is optimized "
            "for small or large sites. In environments with a "
            "lot of users it can be very slow or impossible to "
            "build a list all users. This option tunes the user "
            "interface and behaviour of Plone for this case by "
            "allowing you to search for users instead of "
            "listing all of them."
        ),
        default=False,
        required=False,
    )


def validate_twitter_username(value):
    if value and value.startswith("@"):
        raise Invalid('Twitter username should not include the "@" prefix character.')
    return True


class ISocialMediaSchema(Interface):

    share_social_data = schema.Bool(
        title=_("Share social data"),
        description=_(
            "Include meta tags on pages to give hints to "
            "social media on how to better render your pages "
            "when shared"
        ),
        default=True,
        required=False,
    )

    twitter_username = schema.ASCIILine(
        title=_("Twitter username"),
        description=_(
            "To identify things like Twitter Cards. "
            'Do not include the "@" prefix character.'
        ),
        required=False,
        default="",
        constraint=validate_twitter_username,
    )

    facebook_app_id = schema.ASCIILine(
        title=_("Facebook App ID"),
        description=_("To be used with some integrations like Open Graph data"),
        required=False,
        default="",
    )

    facebook_username = schema.ASCIILine(
        title=_("Facebook username"),
        description=_("For linking Open Graph data to a Facebook account"),
        required=False,
        default="",
    )


class IImagingSchema(Interface):
    allowed_sizes = schema.List(
        title=_("Allowed image sizes"),
        description=_(
            "Specify all allowed maximum image dimensions, one per line. The "
            "required format is &lt;name&gt; &lt;width&gt;:&lt;height&gt;."
        ),
        value_type=schema.TextLine(),
        default=[
            "huge 1600:65536",
            "great 1200:65536",
            "larger 1000:65536",
            "large 800:65536",
            "teaser 600:65536",
            "preview 400:65536",
            "mini 200:65536",
            "thumb 128:128",
            "tile 64:64",
            "icon 32:32",
            "listing 16:16",
        ],
        missing_value=[],
        required=False,
    )

    quality = schema.Int(
        title=_("Scaled image quality"),
        description=_(
            "A value for the quality of scaled images, from 1 "
            "(lowest) to 95 (highest). A value of 0 will mean "
            "plone.scaling's default will be used, which is "
            "currently 88."
        ),
        min=0,
        max=95,
        default=88,
    )

    highpixeldensity_scales = schema.Choice(
        title=_("High pixel density mode"),
        description=_(""),
        default="disabled",
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("disabled", "disabled", "Disabled"),
                SimpleTerm("2x", "2x", "Enabled (2x)"),
                SimpleTerm("3x", "3x", "Enabled (2x, 3x)"),
            ]
        ),
    )

    quality_2x = schema.Int(
        title=_("Image quality at 2x"),
        description=_(
            "A value for the quality of 2x high pixel density images, from 1 "
            "(lowest) to 95 (highest). A value of 0 will mean "
            "plone.scaling's default will be used, which is "
            "currently 62."
        ),
        min=0,
        max=95,
        default=62,
    )

    quality_3x = schema.Int(
        title=_("Image quality at 3x"),
        description=_(
            "A value for the quality of 3x high pixel density images, from 1 "
            "(lowest) to 95 (highest). A value of 0 will mean "
            "plone.scaling's default will be used, which is "
            "currently 51."
        ),
        min=0,
        max=95,
        default=51,
    )

    picture_variants = schema.JSONField(
        title=_("Picture variants"),
        description=_("Enter a JSON-formatted picture variants configuration."),
        schema=json.dumps(
            {
                "title": "Image srcset defintion",
                "type": "object",
                "additionalProperties": {"$ref": "#/$defs/srcset"},
                "$defs": {
                    "srcset": {
                        "type": "object",
                        "properties": {
                            "title": {
                                "type": "string",
                            },
                            "preview": {
                                "type": "string",
                            },
                            "hideInEditor": {
                                "type": "boolean",
                            },
                            "sourceset": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "scale": {
                                            "type": "string",
                                        },
                                        "media": {
                                            "type": "string",
                                        },
                                        "additionalScales": {
                                            "type": "array",
                                        },
                                    },
                                    "additionalProperties": False,
                                    "required": ["scale"],
                                },
                            },
                        },
                        "additionalProperties": False,
                        "required": ["title", "sourceset"],
                    },
                },
            }
        ),
        default={
            "large": {
                "title": "Large",
                "sourceset": [
                    {
                        "scale": "larger",
                        "additionalScales": ["preview", "teaser", "large", "great", "huge"],
                    },
                ],
            },
            "medium": {
                "title": "Medium",
                "sourceset": [
                    {
                        "scale": "teaser",
                        "additionalScales": ["preview", "large", "larger", "great"],
                    },
                ],
            },
            "small": {
                "title": "Small",
                "sourceset": [
                    {
                        "scale": "preview",
                        "additionalScales": ["large", "larger"],
                    },
                ],
            },
        },
        required=True,
    )

    image_captioning = schema.Bool(
        title=_("image_captioning_title", "Enable image captioning"),
        description=_(
            "image_captioning_description",
            "Enable automatic image captioning for images set in the richtext editor based on the description of images.",
        ),
        default=True,
        required=False,
    )


class ILoginSchema(Interface):

    auth_cookie_length = schema.Int(
        title=_("Auth cookie length"), default=0, required=False
    )

    verify_login_name = schema.Bool(
        title=_("Verify login name"), default=True, required=False
    )

    allow_external_login_sites = schema.Tuple(
        title=_("Allow external login sites"),
        default=(),
        value_type=schema.ASCIILine(),
        required=False,
    )

    external_login_url = schema.ASCIILine(
        title=_("External login url"), default=None, required=False
    )

    external_logout_url = schema.ASCIILine(
        title=_("External logout url"), default=None, required=False
    )

    external_login_iframe = schema.Bool(
        title=_("External login iframe"), default=False, required=False
    )


class ILinkSchema(Interface):

    external_links_open_new_window = schema.Bool(
        title=_("Open external links in new a window"),
        description=_(""),
        default=False,
        required=False,
    )

    mark_special_links = schema.Bool(
        title=_("Mark special links"),
        description=_("Marks external or special protocol links with class."),
        default=False,
        required=False,
    )


def _check_tales_expression(value):
    from Products.PageTemplates.Expressions import getEngine

    try:
        getEngine().compile(value)
    except Exception:
        raise Invalid(
            _(
                'The expression "${value}" is invalid',
                mapping={"value": value},
            )
        )
    return True


class IActionSchema(Interface):

    category = schema.Choice(
        title=_("Category"),
        vocabulary="plone.app.vocabularies.PortalActionCategories",
        required=True,
    )

    title = schema.TextLine(title=_("Title"), required=True)

    description = schema.Text(title=_("Description"), required=False)

    i18n_domain = schema.TextLine(
        title=_("i18n_domain_heading", default="I18n domain"),
        default="plone",
        required=False,
    )

    url_expr = schema.ASCIILine(
        title=_("action_url_heading", default="Action URL"),
        description=_(
            "action_url_description",
            default="An expression producing the called URL. "
            "Example: string:${globals_view/navigationRootUrl}/page",
        ),
        required=True,
        constraint=_check_tales_expression,
    )

    available_expr = schema.ASCIILine(
        title=_("action_condition_heading", default="Condition"),
        description=_("action_condition_description", default="A boolean expression"),
        required=False,
    )

    permissions = schema.List(
        title=_("action_permissions_heading", default="Permissions"),
        required=True,
        default=["View"],
        missing_value=[],
        value_type=schema.Choice(vocabulary="plone.app.vocabularies.Permissions"),
    )

    visible = schema.Bool(
        title=_("action_visibility_heading", default="Visible?"),
        default=True,
        required=False,
    )

    position = schema.Int(
        title=_("action_position_heading", default="Position"),
        default=1,
        min=1,
        required=True,
    )


class INewActionSchema(Interface):

    category = schema.Choice(
        title=_("Category"),
        vocabulary="plone.app.vocabularies.PortalActionCategories",
        required=True,
    )

    id = schema.ASCIILine(title=_("Id"), required=True)

    @invariant
    def validate_category_id(data):
        categoryid = data.category
        pa = getToolByName(getSite(), "portal_actions")
        category = pa.get(categoryid, {})
        actionid = data.id
        if actionid in category:
            raise Invalid(
                _(
                    'An action with the id "${actionid}" already exists',
                    mapping={"actionid": actionid},
                )
            )
        try:
            category._checkId(actionid)
        except Exception:
            raise Invalid(
                _(
                    'The id "${actionid}" is invalid',
                    mapping={"actionid": actionid},
                )
            )


class IPloneControlPanelView(Interface):
    """A marker interface for views showing a controlpanel."""


class IPloneControlPanelForm(IPloneControlPanelView):
    """Forms using plone.app.controlpanel"""

    def _on_save():
        """Callback mehod which can be implemented by control panels to
        react when the form is successfully saved. This avoids the need
        to re-define actions only to do some additional notification or
        configuration which cannot be handled by the normal schema adapter.

        By default, does nothing.
        """


class IConfigurationChangedEvent(Interface):
    """An event which is fired after a configuration setting has been changed."""

    context = Attribute("The configuration context which was changed.")

    data = Attribute("The configuration data which was changed.")



from zope.interface import Attribute
from zope.interface import Interface


class IPasswordResetToolView(Interface):
    """BrowserView with utility methods"""

    def encode_mail_header(text):
        """Encodes text into correctly encoded email header"""

    def encoded_mail_sender():
        """returns encoded version of Portal name <portal_email>"""

    def registered_notify_subject():
        """returns encoded version of registered notify template subject line"""

    def mail_password_subject():
        """returns encoded version of mail password template subject line"""


class IPWResetTool(Interface):
    """Defines an interface for a tool that provides a facility to
    reset forgotten passwords.

    This interface is rather sparse, but sufficient to describe the
    task. (In this manner we void being dependant on a specific
    process) The details of the process are in the implementation,
    where they belong."""

    id = Attribute("id", 'Must be set to "portal_password_reset"')

    def requestReset(userid):
        """Ask the system to start the password reset procedure for
        user 'userid'.

        Returns the random string that must be used to reset the
        password."""

    def resetPassword(userid, randomstring, password):
        """Set the password (in 'password') for the user who maps to
        the string in 'randomstring'. The 'userid' parameter is provided
        in case extra authentication is needed."""



from zope.interface import Interface


class INonStructuralFolder(Interface):
    """Marker for folderish content types that are folderish as an
    implementation detail only.

    By declaring support for this interface, a content type will not be
    considered folderish by the catalog's is_folderish index/metadata, meaning
    that it will not be treated as folderish by the navigation tree, portal tab
    generation and folder_contents.
    """



from zope.interface import Interface


class ITranslationServiceTool(Interface):
    """Utility methods to access the translation machinery"""

    def translate(*args, **kw):
        """Translate method to access the translation service
        from resticted code like skins.
        """

    def encode(m, input_encoding=None, output_encoding=None, errors="strict"):
        """Encode a give unicode type or string type to string type in encoding
        output_encoding
        """

    def asunicodetype(m, input_encoding=None, errors="strict"):
        """Create type unicode from type string"""

    def ulocalized_time(
        time, long_format=None, time_only=None, context=None, domain="plonelocales"
    ):
        """Returns localized time."""

    def day_msgid(number, format=""):
        """Returns the msgid which can be passed to the translation service for
        l10n of weekday names. Format is either '', 'a' or 's'.
        """

    def month_msgid(number, format=""):
        """Returns the msgid which can be passed to the translation service for
        l10n of month names. Format is either '' or 'a' (long or
        abbreviation).
        """

    def month_english(number, format=""):
        """Returns the english name of month by number. Format is either '' or
        'a' (long or abbreviation).
        """

    def weekday_english(number, format=""):
        """Returns the english name of a week by number. Format is either '',
        'a' or 'p'.
        """



from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.base.interfaces.defaultpage import IDefaultPage
from plone.registry.interfaces import IRegistry
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2Base
from Products.CMFCore.interfaces import IFolderish
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFDynamicViewFTI.interfaces import IBrowserDefault
from Products.CMFDynamicViewFTI.interfaces import IDynamicViewTypeInformation
from Products.Five.browser import BrowserView
from zope.component import getUtility
from zope.component import queryAdapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer


def get_default_page(context):
    """Given a folderish item, find out if it has a default-page using
    the following lookup rules:

        1. A content object called 'index_html' wins
        2. Else check for IBrowserDefault, either if the container implements
           it or if an adapter exists. In both cases fetch its FTI and either
           take it if it implements IDynamicViewTypeInformation or adapt it to
           IDynamicViewTypeInformation. call getDefaultPage on the implementer
           and take value if given.
        3. Else, look up the attribute default_page on the object, without
           acquisition in place
        3.1 look for a content in the container with the id, no acquisition!
        3.2 look for a content at portal, with acquisition
        4. Else, look up the property default_page in site_properties for
           magic ids and test these

    The id of the first matching item is then used to lookup a translation
    and if found, its id is returned. If no default page is set, None is
    returned. If a non-folderish item is passed in, return None always.
    """
    # met precondition?
    if not IFolderish.providedBy(context):
        return

    # The ids where we look for default - must support __contains__
    ids = set()

    # For BTreeFolders we just use the __contains__ otherwise build a set
    if isinstance(aq_base(context), BTreeFolder2Base):
        ids = context
    elif hasattr(aq_base(context), "objectIds"):
        ids = set(context.objectIds())

    # 1. test for contentish index_html
    if "index_html" in ids:
        return "index_html"

    # 2. Test for IBrowserDefault
    if IBrowserDefault.providedBy(context):
        browserDefault = context
    else:
        browserDefault = queryAdapter(context, IBrowserDefault)

    if browserDefault is not None:
        fti = context.getTypeInfo()
        if fti is not None:
            if IDynamicViewTypeInformation.providedBy(fti):
                dynamic_fti = fti
            else:
                dynamic_fti = queryAdapter(fti, IDynamicViewTypeInformation)
            if dynamic_fti is not None:
                page = dynamic_fti.getDefaultPage(context, check_exists=True)
                if page is not None:
                    return page

    # 3.1 Test for default_page attribute in folder, no acquisition
    pages = getattr(aq_base(context), "default_page", [])
    if isinstance(pages, str):
        pages = [pages]
    for page in pages:
        if page and page in ids:
            return page

    portal = queryUtility(ISiteRoot)
    # Might happen during portal creation
    if portal is None:
        return

    # 3.2 Test for default page in portal, acquire
    for page in pages:
        if portal.unrestrictedTraverse(page, None):
            return page

    # 4. Test for default sitewide default_page setting
    registry = getUtility(IRegistry)
    for page in registry.get("plone.default_page", []):
        if page in ids:
            return page


def is_default_page(container, obj):
    """Finds out if the given obj is the default page in its parent folder.

    Only considers explicitly contained objects, either set as index_html,
    with the default_page property, or using IBrowserDefault.
    """
    parent_default_page = get_default_page(container)
    precondition = (
        parent_default_page is not None
        and "/" not in parent_default_page
        and hasattr(aq_base(obj), "getId")
    )
    return precondition and (parent_default_page == obj.getId())


@implementer(IDefaultPage)
class DefaultPageView(BrowserView):
    def isDefaultPage(self, obj):
        return is_default_page(aq_inner(self.context), obj)

    def getDefaultPage(self):
        return get_default_page(aq_inner(self.context))


def _getDefaultPageView(obj, request):
    """This is a nasty hack because the view lookup fails when it occurs too
    early in the publishing process because the request isn't marked with
    the default skin.  Explicitly marking the request appears to cause
    connection errors, so we just instantiate the view manually.
    """
    view = queryMultiAdapter((obj, request), name="default_page")
    if view is None:
        view = DefaultPageView(obj, request)
    return view


def check_default_page_via_view(obj, request):
    container = aq_parent(aq_inner(obj))
    if container is None:
        return False
    view = _getDefaultPageView(container, request)
    return view.isDefaultPage(obj)


def get_default_page_via_view(obj, request):
    # Short circuit if we are not looking at a Folder
    if not obj.isPrincipiaFolderish:
        return None
    view = _getDefaultPageView(obj, request)
    return view.getDefaultPage()



from zope.interface import Interface


class IATDProxyView(Interface):
    """Proxy view for the 'After the Deadline" spellchecker"""

    def checkDocument(self):
        """Proxy for the AtD service's checkDocument function
        See http://www.afterthedeadline.com/api.slp for more info.
        """



from ZTUtils.Lazy import LazyMap

import unittest


class TestBatch(unittest.TestCase):
    def test_batch_no_lazy(self):
        from plone.base.batch import Batch

        batch = Batch(range(100), size=10, start=10)
        self.assertEqual([b for b in batch], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19])

    def test_batch_lazy_map(self):
        from plone.base.batch import Batch

        def get(key):
            return key

        sequence = LazyMap(get, range(80, 90), actual_result_count=95)
        batch = Batch(sequence, size=10, start=80)
        self.assertEqual([b for b in batch], [80, 81, 82, 83, 84, 85, 86, 87, 88, 89])

        self.assertEqual(batch.numpages, 10)
        self.assertEqual(batch.pagenumber, 9)
        self.assertEqual(batch.navlist, range(6, 11))
        self.assertEqual(batch.leapback, [])
        self.assertEqual(batch.prevlist, range(6, 9))
        self.assertEqual(batch.previous.length, 10)
        self.assertEqual(batch.next.length, 5)
        self.assertEqual(batch.pageurl({}), "b_start:int=80")
        self.assertListEqual(
            list(batch.prevurls({})),
            [
                (6, "b_start:int=50"),
                (7, "b_start:int=60"),
                (8, "b_start:int=70"),
            ],
        )
        self.assertListEqual(
            list(batch.nexturls({})),
            [(10, "b_start:int=90")],
        )



from zope.interface import Interface


ACQUIRE = -1  # acquire locallyAllowedTypes from parent (default)
DISABLED = 0  # use default behavior of PortalFolder,
# which uses the FTI information
ENABLED = 1  # allow types from locallyAllowedTypes only


class IConstrainTypes(Interface):
    """
    Interface for folderish content types supporting restricting addable types
    on a per-instance basis.
    """

    def getConstrainTypesMode():
        """
        Find out if add-restrictions are enabled. Returns 0 if they are
        disabled (the type's default FTI-set allowable types is in effect),
        1 if they are enabled (only a selected subset if allowed types will be
        available), and -1 if the allowed types should be acquired from the
        parent. Note that in this case, if the parent portal type is not the
        same as the portal type of this object, fall back on the default (same
        as 0)
        """

    def getLocallyAllowedTypes():
        """
        Get the list of FTI ids for the types which should be allowed to be
        added in this container.
        """

    def getImmediatelyAddableTypes():
        """
        Return a subset of the FTI ids from getLocallyAllowedTypes() which
        should be made most easily available.
        """

    def getDefaultAddableTypes():
        """
        Return a list of FTIs which correspond to the list of FTIs available
        when the constraint mode = 0 (that is, the types addable without any
        setLocallyAllowedTypes trickery involved)
        """

    def allowedContentTypes():
        """
        Return the list of currently permitted FTIs.
        """


class ISelectableConstrainTypes(IConstrainTypes):
    """
    Extension to the IConstrainTypes interface for content types which allow
    the user to set the allowable content types and immediately available
    types.
    """

    def setConstrainTypesMode(mode):
        """
        Set how addable types is controlled in this class. If mode is 0, use
        the type's default FTI-set allowable types). If mode is 1, use only
        those types explicitly enabled using setLocallyAllowedTypes(). If
        mode is -1, acquire the allowable types from the parent. If the parent
        portal type is not the same as this object's type, fall back on the
        behaviour obtained if mode == 0.
        """

    def setLocallyAllowedTypes(types):
        """
        Set a list of type ids which should be allowed. This must be a
        subset of the type's FTI-set allowable types. This list only comes
        into effect when the restrictions mode is 1 (enabled).
        """

    def setImmediatelyAddableTypes(types):
        """
        Set the list of type ids which should be immediately/most easily
        addable. This list must be a subset of any types set in
        setLocallyAllowedTypes.
        """

    def canSetConstrainTypes():
        """
        Return True if the current user is permitted to constrain addable
        types in this folderish object.
        """



# flake8: noqa
from .basetool import IPloneBaseTool
from .basetool import IPloneCatalogTool
from .basetool import IPloneTool
from .breadcrumbs import IHideFromBreadcrumbs
from .constrains import IConstrainTypes
from .constrains import ISelectableConstrainTypes
from .controlpanel import IActionSchema
from .controlpanel import IControlPanel
from .controlpanel import IDateAndTimeSchema
from .controlpanel import IEditingSchema
from .controlpanel import IFilterSchema
from .controlpanel import IImagingSchema
from .controlpanel import ILinkSchema
from .controlpanel import ILoginSchema
from .controlpanel import IMailSchema
from .controlpanel import IMaintenanceSchema
from .controlpanel import IMarkupSchema
from .controlpanel import INavigationSchema
from .controlpanel import INewActionSchema
from .controlpanel import ISearchSchema
from .controlpanel import ISecuritySchema
from .controlpanel import ISiteSchema
from .controlpanel import ISocialMediaSchema
from .controlpanel import ITinyMCEAdvancedSchema
from .controlpanel import ITinyMCELayoutSchema
from .controlpanel import ITinyMCEPluginSchema
from .controlpanel import ITinyMCEResourceTypesSchema
from .controlpanel import ITinyMCESchema
from .controlpanel import ITinyMCESpellCheckerSchema
from .controlpanel import ITypesSchema
from .controlpanel import IUserGroupsSettingsSchema
from .events import IConfigurationChangedEvent
from .events import IReorderedEvent
from .events import ISiteManagerCreatedEvent
from .images import IImageScalesAdapter
from .images import IImageScalesFieldAdapter
from .installable import INonInstallable
from .interface import IInterfaceTool
from .language import ILanguage
from .login import IForcePasswordChange
from .login import IInitialLogin
from .login import ILogin
from .login import ILoginForm
from .login import ILoginFormSchema
from .login import ILoginHelpForm
from .login import ILoginHelpFormSchema
from .login import IRedirectAfterLogin
from .migration import IMigrationTool
from .password_reset import IPasswordResetToolView
from .password_reset import IPWResetTool
from .patterns import IPatternsSettings
from .properties import IPropertiesTool
from .properties import ISimpleItemWithProperties
from .resources import IBundleRegistry
from .resources import IResourceRegistry
from .siteroot import IMigratingPloneSiteRoot
from .siteroot import IPloneSiteRoot
from .siteroot import ITestCasePloneSiteRoot
from .structure import INonStructuralFolder
from .syndication import ISiteSyndicationSettings
from .translationservice import ITranslationServiceTool
from .workflow import IWorkflowChain



from zope.interface import Attribute
from zope.interface import Interface
from zope.interface.interfaces import IObjectEvent


class ISiteManagerCreatedEvent(IObjectEvent):
    """An event that's fired once the Plone portal is enabled as a site."""


class IReorderedEvent(IObjectEvent):
    """An event that's fired once the Plone Tool has been notified of
    a reordering
    """


class IConfigurationChangedEvent(Interface):
    """An event which is fired after a configuration setting has been changed."""

    context = Attribute("The configuration context which was changed.")

    data = Attribute("The configuration data which was changed.")



from plone.app.layout.navigation.interfaces import INavigationRoot
from Products.CMFCore.interfaces import ISiteRoot
from zope.interface import Interface


class IPloneSiteRoot(ISiteRoot, INavigationRoot):
    """
    Marker interface for the object which serves as the root of a
    Plone site.
    """


class IMigratingPloneSiteRoot(Interface):
    """
    Marker interface used for migration GenericSetup profiles.
    """


class ITestCasePloneSiteRoot(Interface):
    """
    Marker interface used for test fixture GenericSetup profiles.
    """



from zope.interface import Interface


class IDefaultPage(Interface):
    """Interface for a view that can determine if its context is the
    default page or not.
    """

    def isDefaultPage(obj):
        """Finds out if the given obj is the default page for the
        adapted object.
        """

    def getDefaultPage():
        """Returns the id of the default page for the adapted object."""



from zope.interface import Interface


class IImageScalesAdapter(Interface):
    """Return a list of image scales for the given context."""

    def __init__(context, request):
        """Adapts context and the request."""

    def __call__():
        """Call IImageScalesFieldAdapter on all fields."""


class IImageScalesFieldAdapter(Interface):
    """Adapter from field to image_scales.

    This is called by an IImageScalesAdapter.
    Default expectation is that there will be adapters for image fields
    and not for others.  But adapters for text fields or relation fields
    are imaginable.
    """

    def __init__(field, context, request):
        """Adapts field, context and request."""

    def __call__():
        """Returns JSON compatible python data."""



""" Unit tests for plone.base.i18nl10n module. """
from contextlib import contextmanager
from unittest.mock import MagicMock
from unittest.mock import patch
from zope.publisher.browser import TestRequest

import datetime
import DateTime
import locale
import unittest


class DummyContext:
    """Dummy context with only a REQUEST.

    The ulocalized_time method needs a context, for example the Plone portal,
    but it is only used to get the request:

    request = aq_acquire(context, "REQUEST")
    """

    def __init__(self):
        self.REQUEST = TestRequest()


@contextmanager
def patch_formatstring(value=None):
    import plone.base.i18nl10n

    with patch.object(
        plone.base.i18nl10n, "get_formatstring_from_registry", return_value=value
    ):
        yield


# Mock translations
TRANSLATIONS = {
    "de": {
        "date_format_short": "DE: ${d}.${m}.${Y}",
    },
    "nl": {
        "date_format_short": "NL: ${a} ${d} ${b} ${Y}",
        "date_format_long": "NL: ${A} ${d} ${B} ${Y}",
        "weekday_sun": "zondag",
        "weekday_sun_abbr": "zo",
        "month_mar": "maart",
        "month_mar_abbr": "mrt",
    },
    None: {
        "date_format_short": "${b} ${d}, ${Y}",
        "date_format_long": "${b} ${d}, ${Y} ${I}:${M} ${p}",
        "time_format": "${I}:${M} ${p}",
        "weekday_sun": "Sunday",
        "weekday_sun_abbr": "Su",
        "month_mar": "March",
        "month_mar_abbr": "Mar",
    },
}


def mock_translate(msgid, *args, **kwargs):
    from zope.i18n import translate

    target_language = kwargs.get("target_language")
    default = kwargs.get("default")
    override = False
    # Note: we have translations for target_language=None .
    try:
        override = TRANSLATIONS[target_language][msgid]
    except Exception:
        pass
    # Even if the TRANSLATIONS lookup worked, we may still need to call the
    # original translate function.  This depends on the keyword arguments having
    # a mapping and/or a default.
    if override:
        orig_msgid = msgid
        msgid = override
    standard = translate(msgid, *args, **kwargs)
    if standard == default and override:
        # Example: original msgid is "weekday_sun", just like the default,
        # and this is the standard answer.  But override is "zondag".
        # The override should win then.
        return override
    return standard


@contextmanager
def patch_translate():
    import plone.base.i18nl10n

    with patch.object(plone.base.i18nl10n, "translate", wraps=mock_translate):
        yield


@contextmanager
def use_locale(value=None):
    orig = locale.getlocale(locale.LC_TIME)[0] or "C"
    try:
        worked = locale.setlocale(locale.LC_TIME, value)
    except locale.Error:
        worked = False
    yield worked
    locale.setlocale(locale.LC_TIME, orig)


class BasicI18nl10nTests(unittest.TestCase):
    def test_regexp_dt_format_string_regexp(self):
        from plone.base.i18nl10n import _dt_format_string_regexp

        dt_string = "%Y-%m-%d %H:%M"
        locales_string = "${H}:${M}"

        # test for strftime format string
        self.assertTrue(bool(_dt_format_string_regexp.findall(dt_string)))
        self.assertFalse(bool(_dt_format_string_regexp.findall(locales_string)))

    def test_regexp_interp_regex(self):
        from plone.base.i18nl10n import _interp_regex

        locales_string = "${H}:${M}"

        # test for locale string elements:
        self.assertEqual(
            _interp_regex.findall(locales_string),
            ["${H}", "${M}"],
        )

    def test_utranslate(self):
        from plone.base.i18nl10n import utranslate

        request = TestRequest()

        # Test string value
        value = utranslate("domain", "foo", context=request)
        self.assertEqual(value, "foo")

        # Test empty string
        value = utranslate("domain", "", context=request)
        self.assertEqual(value, "")

        # Test empty domain
        value = utranslate("", "foo", context=request)
        self.assertEqual(value, "foo")

        # Test default is None
        value = utranslate("domain", "foo", context=request, default=None)
        self.assertEqual(value, "foo")

        # Test default is other
        value = utranslate("domain", "foo", context=request, default="other")
        self.assertEqual(value, "other")

        # Test a target language.
        value = utranslate("domain", "foo", context=request, target_language="nl")
        self.assertEqual(value, "foo")

    def test_ulocalized_time_fallbacks(self):
        """Test the fallbacks in the ulocalized_time method.

        The ulocalized_time method uses one of three formats, based on the arguments:

        if time_only:
            msgid = "time_format"  # fallback: %H:%M
        elif long_format:
            msgid = "date_format_long"  # fallback: %Y-%m-%d %H:%M
        else:
            msgid = "date_format_short"  # fallback: %Y-%m-%d

        The ulocalized_time method looks this msgid up in a po file to see which format
        to use when calling 'strftime'.  In these unit tests, no translations are
        available, so the hardcoded fallbacks are used.

        Default Plone site has a few records with which you can override this.
        See get_formatstring_from_registry.  In plain unit tests this call fails,
        because no registry has been set up.

        Easiest way for us to test this, is then by mocking/patching this function.
        To test the fallbacks, we only have to use None as return value.
        """
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            # We test all combinations of long_format and time_only,
            # although not all make sense (both true).
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    long_format=True,
                    time_only=False,
                    context=context,
                ),
                "1997-03-09 13:45",
            )
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    long_format=False,
                    time_only=False,
                    context=context,
                ),
                "1997-03-09",
            )
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    long_format=False,
                    time_only=True,
                    context=context,
                ),
                "13:45",
            )
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    long_format=False,
                    time_only=True,
                    context=context,
                ),
                "13:45",
            )

    def test_ulocalized_time_formats(self):
        """Test different formats in the ulocalized_time method.

        See notes in ulocalized_time_fallbacks.  Best way to check non-default formats,
        is to patch get_formatstring_from_registry.  This results in ignoring the
        time_only and long_format options, so we do not pass them here.
        """
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring("%H:%M %d-%m-%Y"):
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    context=context,
                ),
                "13:45 09-03-1997",
            )
        with patch_formatstring("German: %Y.%m.%d"):
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    context=context,
                ),
                "German: 1997.03.09",
            )
        # Use week days and month names.
        with patch_formatstring("%A %B %d %Y"):
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    context=context,
                ),
                "Sunday March 09 1997",
            )
        with patch_formatstring("%a %b %d %Y"):
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    context=context,
                ),
                "Sun Mar 09 1997",
            )

    def test_ulocalized_time_translate_no_language_format(self):
        """Test translations in the ulocalized_time method.

        This is mostly about translating week days and month names.
        But also about using a different format for a language.

        For this test, nothing was changed.
        A few other tests follow, where we did overrides.
        """
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            with patch_translate():
                # French gets the default.
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                        target_language="fr",
                    ),
                    "1997-03-09",
                )

    def test_ulocalized_time_translate_language_format(self):
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            with patch_translate():
                # For German we have mocked a different date_format_short.
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                        target_language="de",
                    ),
                    "DE: 09.03.1997",
                )

    def test_ulocalized_time_translate_weekdays_month_names_short(self):
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            with patch_translate():
                # For Dutch we have mocked a different date_format_short,
                # using abbreviations for week days and month names.
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                        target_language="nl",
                    ),
                    "NL: zo 09 mrt 1997",
                )

    def test_ulocalized_time_translate_weekdays_month_names_long(self):
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            with patch_translate():
                # For Dutch we have mocked a different date_format_long,
                # using week days and month names.
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        long_format=True,
                        context=context,
                        target_language="nl",
                    ),
                    "NL: zondag 09 maart 1997",
                )

    def test_ulocalized_time_dollar_classic_format_english(self):
        # When the format is a classic strftime format as you would use in standard
        # Python, the week days and month names are not translated.
        # Python will do translation using the current locale.
        # In these tests, it is useless to pass a target_language.

        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring("%A %B %d %Y"):
            # Default is English, which needs no translation.
            # Do not assume that English is the locale on the system though.
            with use_locale("C") as available:
                if not available:
                    self.skipTest("English (C) locale not available")
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                    ),
                    "Sunday March 09 1997",
                )

    def test_ulocalized_time_dollar_classic_format_dutch(self):
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring("%A %B %d %Y"):
            # But how about Dutch?  This can only be tested if we have the this locale
            # and set it here.
            with use_locale("nl_NL") as available:
                if not available:
                    self.skipTest("Dutch (nl_NL) locale not available")
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                    ),
                    "zondag maart 09 1997",
                )

    def test_ulocalized_time_dollar_format_translates_weekdays_month_names_long(self):
        # When the format is taken from the registry instead of from po files,
        # the week days and month names should still be translated.
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring("${A} ${B} ${d} ${Y}"):
            with patch_translate():
                # English should be no trouble, because it needs no translation.
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                        # target_language="en",
                    ),
                    "Sunday March 09 1997",
                )
                # But how about Dutch?  In the tests we have Dutch translations for
                # Sunday and March, and they should be used.
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        context=context,
                        target_language="nl",
                    ),
                    "zondag maart 09 1997",
                )

    def test_ulocalized_time_explicit_long_format_dollar(self):
        # We took over a change from experimental.ulocalized_time:
        # allow passing an explicit long_format.
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            with patch_translate():
                self.assertEqual(
                    ulocalized_time(
                        "Mar 9, 1997 1:45pm",
                        long_format="NL: ${A} go ${d} crazy ${B} ${Y}!",
                        context=context,
                        target_language="nl",
                    ),
                    "NL: zondag go 09 crazy maart 1997!",
                )

    def test_ulocalized_time_explicit_long_format_classic(self):
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()
        with patch_formatstring():
            with patch_translate():
                # You should not use %, but it can work.
                with use_locale("nl_NL") as available:
                    if not available:
                        self.skipTest("Dutch (nl_NL) locale not available")
                    self.assertEqual(
                        ulocalized_time(
                            "Mar 9, 1997 1:45pm",
                            long_format="NL: %A go %d crazy %B %Y!",
                            context=context,
                            # this is ignored:
                            target_language="de",
                        ),
                        "NL: zondag go 09 crazy maart 1997!",
                    )

    def test_ulocalized_time_no_context(self):
        # Without context, we fall back to ISO8601.
        from plone.base.i18nl10n import ulocalized_time

        self.assertEqual(
            ulocalized_time("Mar 9, 1997 1:45pm", context=None), "1997-03-09T13:45:00"
        )

    def test_ulocalized_time_none(self):
        # Test passing None.
        from plone.base.i18nl10n import ulocalized_time

        context = DummyContext()

        with patch_formatstring():
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    time_only=None,
                    context=context,
                ),
                "1997-03-09",
            )
            self.assertEqual(
                ulocalized_time(
                    "Mar 9, 1997 1:45pm",
                    time_only=None,
                    domain=None,
                    context=context,
                ),
                "1997-03-09",
            )

    def test_ulocalized_time_fetch_error(self):
        from plone.base.i18nl10n import ulocalized_time

        # You should pass a DateTime, although the code does try to convert.
        # If this fails, get get None back.  Apparently in an old bug, an error
        # got passed like this:
        self.assertEqual(ulocalized_time("(Missing.Value,), {}"), None)
        self.assertEqual(ulocalized_time("Hello world"), None)
        self.assertEqual(ulocalized_time([]), None)

    def test_ulocalized_time_conversion(self):
        from plone.base.i18nl10n import ulocalized_time

        # You should pass a DateTime, but when the argument is accepted by DateTime,
        # conversion automatically happens.  When it depends on the time of day,
        # or on the time zone, we cannot compare the full result.
        now = DateTime.DateTime()
        day = now.strftime("%Y-%m-%d")
        self.assertTrue(ulocalized_time(None).startswith(day))
        self.assertTrue(ulocalized_time(datetime.datetime.now()).startswith(day))
        self.assertTrue(ulocalized_time(86400).startswith("1970-01"))
        self.assertEqual(ulocalized_time("Jan 27 1976"), "1976-01-27T00:00:00")



from .. import PloneMessageFactory as _
from zope import schema
from zope.interface import Interface


class ISyndicatable(Interface):
    pass


class IFeedData(Interface):
    def link():
        """
        Link to item
        """

    def base_url():
        """
        base url to item
        """

    def title():
        """
        title of item
        """

    def description():
        """ """

    def categories():
        """
        List of tags
        """

    def published():
        """
        publishing date
        """

    def modified():
        """
        modification date
        """

    def uid():
        """ """

    def rights():
        """ """

    def publisher():
        """ """

    def author():
        """ """

    def author_name():
        """ """

    def author_email():
        """ """


class IFeed(IFeedData):
    """
    An adapter on the context and request
    to get feed information
    """

    def show_about():
        """ """

    def logo():
        """ """

    def icon():
        """ """

    def _brains():
        """
        return list of brains
        """

    def _items():
        """
        return full objects
        """

    def items():
        """
        adapted items
        """

    def limit():
        """ """

    def language():
        """ """


class ISearchFeed(IFeed):
    pass


class IFeedItem(IFeedData):
    """
    An adapter on the feed item and IFeed instance
    """

    def body():
        """ """

    def guid():
        """ """

    def has_enclosure():
        """ """

    def file():
        """ """

    def file_url():
        """ """

    def file_length():
        """ """

    def file_type():
        """ """


class ISiteSyndicationSettings(Interface):

    allowed = schema.Bool(
        title=_("Allowed"),
        description=_("Allow syndication for collections and folders " "on site."),
        default=True,
        required=False
    )

    default_enabled = schema.Bool(
        title=_("Enabled by default"),
        description=_(
            "If syndication should be enabled by default for all "
            "folders and collections."
        ),
        default=False,
        required=False
    )

    search_rss_enabled = schema.Bool(
        title=_("Search RSS enabled"),
        description=_("Allows users to subscribe to feeds of search results"),
        default=True,
        required=False
    )

    show_author_info = schema.Bool(
        title=_("Show author info"),
        description=_("Should feeds include author information"),
        default=True,
        required=False
    )

    render_body = schema.Bool(
        title=_("Render Body"),
        description=_(
            "help_render_body",
            default="If body text available for item, "
            "render it, otherwise use description.",
        ),
        default=False,
        required=False
    )

    max_items = schema.Int(
        title=_("label_syndication_max_items", default="Maximum items"),
        description=_(
            "help_syndication_max_items",
            default="Maximum number of items that will be syndicated.",
        ),
        min=1,
        default=15,
    )

    allowed_feed_types = schema.Tuple(
        title=_("Allowed Feed Types"),
        description=_("Separate view name and title by '|'"),
        required=True,
        missing_value=None,
        default=(
            "RSS|RSS 1.0",
            "rss.xml|RSS 2.0",
            "atom.xml|Atom",
            "itunes.xml|iTunes",
        ),
        value_type=schema.TextLine(),
    )

    site_rss_items = schema.Tuple(
        title=_("Site RSS"),
        description=_(
            "Paths to folders and collections to link to " "at the portal root."
        ),
        required=False,
        default=("/news/aggregator",),
        value_type=schema.Choice(
            vocabulary="plone.app.vocabularies.SyndicatableFeedItems"
        ),
    )

    show_syndication_button = schema.Bool(
        title=_("Show settings button"),
        description=_(
            "Makes it possible to customize syndication settings "
            "for particular folders and collections "
        ),
        default=False,
        required=False
    )

    show_syndication_link = schema.Bool(
        title=_("Show feed link"),
        description=_(
            "Enable RSS link document action on the syndication " "content item."
        ),
        default=False,
        required=False
    )


class IFeedSettings(Interface):

    enabled = schema.Bool(
        title=_("Enabled"),
        default=False,
        required=False
    )

    feed_types = schema.Tuple(
        title=_("Feed Types"),
        required=True,
        missing_value=None,
        default=("RSS", "rss.xml", "atom.xml"),
        value_type=schema.Choice(
            vocabulary="plone.app.vocabularies.SyndicationFeedTypes"
        ),
    )

    render_body = schema.Bool(
        title=_("Render Body"),
        description=_(
            "help_render_body",
            default="If body text available for item, "
            "render it, otherwise use description.",
        ),
        default=False,
        required=False
    )

    max_items = schema.Int(
        title=_("label_syndication_max_items", default="Maximum items"),
        description=_(
            "help_syndication_max_items",
            default="Maximum number of items that will be syndicated.",
        ),
        default=15,
    )


class ISyndicationUtil(Interface):
    def allowed_feed_types():
        """
        get a list of allow feed types
        """

    def context_allowed():
        """
        If syndication is allowed on the context
        """

    def context_enabled(raise404=False):
        """
        If syndication is enabled on the context
        """

    def site_enabled():
        """
        If syndication is enabled on the site
        """

    def search_rss_enabled(raise404=False):
        """
        If search_rss is enabled
        """

    def show_author_info():
        """
        If author information should show on feeds
        """

    def max_items():
        """
        Default max items to show on the site
        """

    def rss_url():
        """
        Default rss url. Mainly to be used for the
        rss portal_action link
        """



from zope.interface import Interface


class IPatternsSettings(Interface):
    """Interface to register global pattern settings adapters"""

    def __call__(self):
        """
        Return a dict of pattern options
        """
        pass



from .. import PloneMessageFactory as _
from plone.schema.email import Email
from plone.z3cform.interfaces import IWrappedForm
from zope import schema
from zope.interface import Interface


class IRedirectAfterLogin(Interface):
    """Redirect after login adapters should provide this interface"""


class IForcePasswordChange(Interface):
    """Hook point to customize forcing a password change"""


class IInitialLogin(Interface):
    """Hook point to customize what happens the first time a user logs into
    the site"""


class ILogin(Interface):
    """Login form schema"""

    login = schema.TextLine(
        title=_("label_log_in", default="Log in"),
    )

    password = schema.Password(
        title=_("label_password", default="Password"),
    )


class ILoginForm(IWrappedForm):
    """Login form marker interface"""


class ILoginFormSchema(Interface):
    """Login form schema"""

    ac_name = schema.TextLine(
        title=_("label_login_name", default="Login Name"),
        required=True,
    )

    ac_password = schema.Password(
        title=_("label_password", default="Password"),
        required=True,
    )

    came_from = schema.TextLine(
        title="Came From",  # not translated, hidden field
        required=False,
    )


class ILoginHelpForm(IWrappedForm):
    """Login Help form marker interface"""


class ILoginHelpFormSchema(Interface):
    """Login Help form schema"""

    reset_password = schema.TextLine(
        title=_("label_pwreset_username", default="Username"),
        description=_(
            "help_pwreset_username",
            default="Enter your username "
            "or email and we’ll send you a password reset link.",
        ),
        required=True,
    )

    recover_username = Email(
        title=_("label_pwreset_email", default="Email"),
        description=_(
            "help_pwreset_email",
            default="Enter your email and " "we’ll send you your username.",
        ),
        required=True,
    )



from zope.interface import Attribute
from zope.interface import Interface


class IPropertiesTool(Interface):

    """Manage properties of the site as a whole."""

    id = Attribute("id", 'Must be set to "portal_properties"')

    def editProperties(props):
        """Change portal settings.

        Permission --  Manage portal
        """

    def smtp_server():
        """Get local SMTP server.

        Returns -- String
        """


class ISimpleItemWithProperties(Interface):
    pass



from plone.batching.batch import QuantumBatch
from plone.batching.utils import calculate_pagerange
from zope.deprecation import deprecate
from ZTUtils import make_query


class Batch(QuantumBatch):

    b_start_str = "b_start"

    def __init__(
        self,
        sequence,
        size,
        start=0,
        end=0,
        orphan=0,
        overlap=0,
        pagerange=7,
        quantumleap=0,
        b_start_str="b_start",
    ):
        super().__init__(
            sequence, size, start, end, orphan, overlap, pagerange, quantumleap
        )
        self.b_start_str = b_start_str

    @deprecate("Use length attribute instead of __len__")
    def __len__(self):
        # Note: Using len() was deprecated for several years.
        # It was recommended to explicitly either use the `length` attribute
        # for the size of the current page, which is what we return now,
        # or use the `sequence_length` attribute for the size of the
        # entire sequence.
        # But the deprecation was reverted for Plone 5.2.3,
        # because core code in Products.PageTemplates called `len`
        # on batches, making the deprecation warning unavoidable
        # and thus unnecessary.
        # See https://github.com/plone/Products.CMFPlone/issues/3176
        return self.length

    def __bool__(self):
        # Without __bool__ a bool(self) would call len(self), which
        # gives a deprecation warning.
        return bool(self.length)

    # For Python 2 compatibility:
    __nonzero__ = __bool__

    def initialize(self, start, end, size):
        super().initialize(start, end, size)
        self.pagerange, self.pagerangestart, self.pagerangeend = calculate_pagerange(
            self.pagenumber, self.numpages, self.pagerange
        )

    def pageurl(self, formvariables, pagenumber=-1):
        # Makes the url for a given page.
        if pagenumber == -1:
            pagenumber = self.pagenumber
        b_start = pagenumber * (self.pagesize - self.overlap) - self.pagesize
        return make_query(formvariables, {self.b_start_str: b_start})

    def navurls(self, formvariables, navlist=None):
        # Returns the page number and url for the navigation quick links.
        if navlist is None:
            navlist = []
        if not navlist:
            navlist = self.navlist
        return map(
            lambda x, formvariables=formvariables: (x, self.pageurl(formvariables, x)),
            navlist,
        )

    def prevurls(self, formvariables):
        # Helper method to get prev navigation list from templates.
        return self.navurls(formvariables, self.previous_pages)

    def nexturls(self, formvariables):
        # Helper method to get next navigation list from templates.
        return self.navurls(formvariables, self.next_pages)

    prevlist = QuantumBatch.previous_pages
    nextlist = QuantumBatch.next_pages



from unittest import TestSuite

import doctest


def test_suite():
    suites = (
        doctest.DocFileSuite(
            "messages.rst",
            package="plone.base.tests",
        ),
    )
    return TestSuite(suites)



from . import PloneMessageFactory as _
from .interfaces import ISearchSchema
from AccessControl import Unauthorized
from Acquisition import aq_base
from Acquisition import aq_get
from Acquisition import aq_parent
from DateTime import DateTime
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import ITypesTool
from Products.CMFCore.utils import getToolByName
from urllib.parse import urlparse
from zExceptions import NotFound
from ZODB.POSException import ConflictError
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.component.interfaces import ISite
from zope.deprecation import deprecate
from zope.i18n import translate
from zope.publisher.interfaces.browser import IBrowserRequest

import logging
import transaction


logger = logging.getLogger("Plone")

SIZE_CONST = {
    "KB": 1024,
    "MB": 1024 ** 2,
    "GB": 1024 ** 3,
    "TB": 1024 ** 4,
    "PB": 1024 ** 5,
}
SIZE_ORDER = ("PB", "TB", "GB", "MB", "KB")

_marker = dict()


def human_readable_size(size):
    """Get a human readable size string."""
    smaller = SIZE_ORDER[-1]

    # if the size is a float, then make it an int
    # happens for large files
    try:
        size = int(size)
    except (ValueError, TypeError):
        pass

    if not size:
        return f"0 {smaller}"

    if not isinstance(size, int):
        return size

    if size < SIZE_CONST[smaller]:
        return f"1 {smaller}"

    for c in SIZE_ORDER:
        if size // SIZE_CONST[c] > 0:
            break
    return f"{float(size / float(SIZE_CONST[c])):.1f} {c}"


def safe_int(value, default=0):
    """Convert value to integer or just return 0 if we can't

    >>> safe_int(45)
    45

    >>> safe_int("42")
    42

    >>> safe_int("spam")
    0

    >>> safe_int([])
    0

    >>> safe_int(None)
    0

    >>> safe_int(None, default=-1)
    -1
    """
    try:
        return int(value)
    except (ValueError, TypeError):
        return default


@deprecate("Use plone.base.utils.safe_int instead (will be removed in Plone 7)")
def safeToInt(value, default=0):
    return safe_int(value, default)


def safe_text(value, encoding="utf-8") -> str:
    """Converts a value to text, even it is already a text string.

    >>> from Products.CMFPlone.utils import safe_text
    >>> test_bytes = u'\u01b5'.encode('utf-8')
    >>> safe_text('spam') == u'spam'
    True
    >>> safe_text(b'spam') == u'spam'
    True
    >>> safe_text(u'spam') == u'spam'
    True
    >>> safe_text(u'spam'.encode('utf-8')) == u'spam'
    True
    >>> safe_text(test_bytes) == u'\u01b5'
    True
    >>> safe_text(u'\xc6\xb5'.encode('iso-8859-1')) == u'\u01b5'
    True
    >>> safe_text(test_bytes, encoding='ascii') == u'\u01b5'
    True
    >>> safe_text(1) == 1
    True
    >>> print(safe_text(None))
    None
    """
    if isinstance(value, bytes):
        try:
            value = str(value, encoding)
        except (UnicodeDecodeError):
            value = value.decode("utf-8", "replace")
    return value


def safe_bytes(value, encoding="utf-8") -> bytes:
    """Convert text to bytes of the specified encoding."""
    if isinstance(value, str):
        value = value.encode(encoding)
    return value


def safe_hasattr(obj, name, _marker=object()):
    """Make sure we don't mask exceptions like hasattr().

    We don't want exceptions other than AttributeError to be masked,
    since that too often masks other programming errors.
    Three-argument getattr() doesn't mask those, so we use that to
    implement our own hasattr() replacement.
    """
    return getattr(obj, name, _marker) is not _marker


def base_hasattr(obj, name):
    """Like safe_hasattr, but also disables acquisition."""
    return safe_hasattr(aq_base(obj), name)


def safe_callable(obj):
    """Make sure our callable checks are ConflictError safe."""
    if safe_hasattr(obj, "__class__"):
        if safe_hasattr(obj, "__call__"):
            return True
        return isinstance(obj, type)
    return callable(obj)


def get_empty_title(context, translated=True):
    """Returns string to be used for objects with no title or id"""
    # The default is an extra fancy unicode elipsis
    empty = b"\x5b\xc2\xb7\xc2\xb7\xc2\xb7\x5d".decode("utf8")
    if translated:
        if context is not None:
            if not IBrowserRequest.providedBy(context):
                context = aq_get(context, "REQUEST", None)
        empty = translate("title_unset", domain="plone", context=context, default=empty)
    return empty


def pretty_title_or_id(context, obj, empty_value=_marker):
    """Return the best possible title or id of an item, regardless
    of whether obj is a catalog brain or an object, but returning an
    empty title marker if the id is not set (i.e. it's auto-generated).
    """
    title = None
    if base_hasattr(obj, "Title"):
        title = getattr(obj, "Title", None)
    if safe_callable(title):
        title = title()
    if title:
        return title
    item_id = getattr(obj, "getId", None)
    if safe_callable(item_id):
        item_id = item_id()
    if item_id is not None:
        return item_id
    if empty_value is _marker:
        empty_value = get_empty_title(context)
    return empty_value


def get_installer(context, request=None):
    if request is None:
        request = aq_get(context, "REQUEST", None)
    view = getMultiAdapter((context, request), name="installer")
    return view


def is_expired(content):
    """Find out if the object is expired (copied from skin script)"""

    expiry = None

    # NOTE: We also accept catalog brains as 'content' so that the
    # catalog-based folder_contents will work. It's a little magic, but
    # it works.

    # ExpirationDate should have an ISO date string, which we need to
    # convert to a DateTime

    # Try DC accessor first
    if base_hasattr(content, "ExpirationDate"):
        expiry = content.ExpirationDate

    # Try the direct way
    if not expiry and base_hasattr(content, "expires"):
        expiry = content.expires

    # See if we have a callable
    if safe_callable(expiry):
        expiry = expiry()

    # Convert to DateTime if necessary, ExpirationDate may return 'None'
    if expiry and expiry != "None" and isinstance(expiry, str):
        expiry = DateTime(expiry)

    if isinstance(expiry, DateTime) and expiry.isPast():
        return 1
    return 0


def get_top_request(request):
    """Get highest request from a subrequest."""

    def _top_request(req):
        parent_request = req.get("PARENT_REQUEST", None)
        return _top_request(parent_request) if parent_request else req

    return _top_request(request)


def get_top_site_from_url(context, request):
    """
    Find the first ISite object that appears in the pre-virtual-hosting URL
    path, falling back to the object pointed by the virtual hosting root.

    Use this method if you need a "root object" reference to generate URLs
    that will be used by, and will work correctly from the standpoint of,
    *browser* code (JavaScript / XML HTTP requests) after virtual hosting has
    been applied.  *Never* use this to get to a site root in Plone server code
    -- it is not appropriate for that use.

    If the current context is within a subsite within a PloneSiteRoot and no
    virtual hosting is in place, the PloneSiteRoot is returned.
    When at the same context but in a virtual hosting environment with the
    virtual host root pointing to the subsite, it returns the subsite instead
    the PloneSiteRoot.  Finally, if the virtual hosting environment points to
    a *child* of a site/subsite, that child returns instead of the site/subsite.

    For this given content structure:

    /Plone/Subsite:
      /file
      /en-US
        /folder
          /image

    It should return the following in these cases:

    - No virtual hosting
      URL path:              /Plone
      Object accessed:       /Plone
      Returns:               Plone

    - No virtual hosting
      URL path:              /Plone/Subsite
      Object accessed:       /Plone/Subsite
      Returns:               Plone

    - Virtual hosting root:  /Plone/Subsite
      URL path:              /
      Object accessed:       /Plone/Subsite
      Returns:               Subsite

    - Virtual hosting root:  /Plone/Subsite
      URL path:              /file
      Object accessd:        /Plone/Subsite/file
      Returns:               Subsite

    - Virtual hosting root:  /Plone/Subsite/en-US
      URL path:              /folder/image
      Object accessed:       /Plone/Subsite/en-US/folder/image
      Returns:               Subsite/en-US
      (in this last case -- common with p.a.multilingual and usually described
       as subdomain hosting for languages -- no Site object is visible TTW,
       so it must return the topmost visible container, since the callees
       need an object with a valid, TTW-visible URL to do their work.)
    """
    site = getSite()
    try:
        # This variable collects all sites found during the traversal that
        # takes place below, which only includes objects visible via VHM.
        subsites = []
        # This variable collect the topmost objects found during the
        # traversal below, as fallback in case there are no sites found
        # during the traversal.
        topmosts = []
        url_path = urlparse(context.absolute_url()).path.split("/")
        for idx in range(len(url_path)):
            _path = "/".join(url_path[: idx + 1]) or "/"
            site_path = "/".join(request.physicalPathFromURL(_path)) or "/"
            # The following line is fine.  We do a restrictedTraverse
            # below to resolve the actual object, so the user (technically,
            # the browser) cannot ever get a reference to an object it does
            # not have permission to.
            try:
                _site = context.unrestrictedTraverse(site_path)
            except NotFound:
                # Oh.  This path is not findable.  So we will not consider
                # it below as a thing we can return to stand in for ISite.
                continue
            if ISite.providedBy(_site):
                subsites.append(idx)
            else:
                topmosts.append(idx)
        # Pick the subsite to return.
        # If no subsite was found, return the top site.
        # If at some point a subsite was found, return the
        # rootmost of all subsites.
        # With VHM, sometimes the topmost site is not actually
        # in the client URL, so in that case we fall back to
        # the topmost accessible object within the client URL.
        try:
            _path_idx = subsites[0]
        except IndexError:
            _path_idx = topmosts[0]
        _path = "/".join(url_path[: _path_idx + 1]) or "/"
        site_path = "/".join(request.physicalPathFromURL(_path)) or "/"
        site = context.restrictedTraverse(site_path)
    except (ValueError, AttributeError) as exc:
        # On error, just return getSite.
        # Refs: https://github.com/plone/plone.app.content/issues/103
        # Also, TestRequest doesn't have physicalPathFromURL
        pass
    return site


def transaction_note(note):
    """Write human legible note"""
    T = transaction.get()
    if (len(T.description) + len(note)) >= 65533:
        logger.warn("Transaction note too large omitting %s" % str(note))
    else:
        T.note(safe_text(note))


def check_id(
    context, id=None, required=0, alternative_id=None, contained_by=None, **kwargs
):
    """Test an id to make sure it is valid.

    This used to be in Products/CMFPlone/skins/plone_scripts/check_id.py.

    Returns an error message if the id is bad or None if the id is good.
    Parameters are as follows:

        id - the id to check

        required - if False, id can be the empty string

        alternative_id - an alternative value to use for the id
        if the id is empty or autogenerated

    Accept keyword arguments for compatibility with the fallback
    in Products.validation.

    Note: The reason the id is included is to handle id error messages for
    such objects as files and images that supply an alternative id when an
    id is auto-generated.
    If you say "There is already an item with this name in this folder"
    for an image that has the Name field populated with an autogenerated id,
    it can cause some confusion; since the real problem is the name of
    the image file name, not in the name of the autogenerated id.
    """

    def xlate(message):
        ts = getToolByName(context, "translation_service", None)
        if ts is None:
            return message
        return ts.translate(message, context=context.REQUEST)

    # if an alternative id has been supplied, see if we need to use it
    if alternative_id and not id:
        id = alternative_id

    # make sure we have an id if one is required
    if not id:
        if required:
            return xlate(_("Please enter a name."))

        # Id is not required and no alternative was specified, so assume the
        # object's id will be context.getId(). We still should check to make
        # sure context.getId() is OK to handle the case of pre-created objects
        # constructed via portal_factory.  The main potential problem is an id
        # collision, e.g. if portal_factory autogenerates an id that already
        # exists.

        id = context.getId()

    #
    # do basic id validation
    #

    # check for reserved names
    if id in (
        "login",
        "layout",
        "plone",
        "zip",
        "properties",
    ):
        return xlate(_("${name} is reserved.", mapping={"name": id}))

    # check for bad characters
    plone_utils = getToolByName(context, "plone_utils", None)
    if plone_utils is not None:
        bad_chars = plone_utils.bad_chars(id)
        if len(bad_chars) > 0:
            bad_chars = "".join(bad_chars).decode("utf-8")
            decoded_id = id.decode("utf-8")
            return xlate(
                _(
                    "${name} is not a legal name. The following characters are "
                    "invalid: ${characters}",
                    mapping={"name": decoded_id, "characters": bad_chars},
                )
            )

    # check for a catalog index
    portal_catalog = getToolByName(context, "portal_catalog", None)
    if portal_catalog is not None:
        if id in list(portal_catalog.indexes()) + list(portal_catalog.schema()):
            return xlate(_("${name} is reserved.", mapping={"name": id}))

    # id is good; decide if we should check for id collisions
    if contained_by is not None:
        # Always check for collisions if a container was passed
        # via the contained_by parameter.
        checkForCollision = True
    else:
        # if we have an existing object, only check for collisions
        # if we are changing the id
        checkForCollision = context.getId() != id

    # check for id collisions
    if not checkForCollision:
        # We are done.
        return
    # handles two use cases:
    # 1. object has not yet been created and we don't know where it will be
    # 2. object has been created and checking validity of id within
    #    container
    if contained_by is None:
        try:
            contained_by = context.getParentNode()
        except Unauthorized:
            return  # nothing we can do
    try:
        result = _check_for_collision(contained_by, id, **kwargs)
    except Unauthorized:
        # There is a permission problem. Safe to assume we can't use this id.
        return xlate(_("${name} is reserved.", mapping={"name": id}))
    if result is not None:
        result = xlate(
            result,
        )
    return result


def _check_for_collision(contained_by, cid, **kwargs):
    """Check for collisions of an object id in a container.

    Accept keyword arguments for compatibility with the fallback
    in Products.validation.

    When this was still a Python skin script, some security checks
    would have been done automatically and caught by some
    'except Unauthorized' lines.  Now, in unrestricted Python
    code, we explicitly check.  But not all checks make sense.  If you don't
    have the 'Access contents information' permission, theoretically
    you should not be able to check for an existing conflicting id,
    but it seems silly to then pretend that there is no conflict.

    For safety, we let the check_id
    function do a try/except Unauthorized when calling us.
    """
    # Check for an existing object.
    if cid in contained_by:
        existing_obj = getattr(contained_by, cid, None)
        if getattr(aq_base(existing_obj), "portal_type", _marker) is not _marker:
            return _(
                "There is already an item named ${name} in this folder.",
                mapping={"name": cid},
            )

    # containers may implement this hook to further restrict ids
    if getattr(aq_base(contained_by), "checkValidId", _marker) is not _marker:
        try:
            contained_by.checkValidId(cid)
        except ConflictError:
            raise
        except:  # noqa: E722
            return _("${name} is reserved.", mapping={"name": cid})

    # make sure we don't collide with any parent method aliases
    types_tool = getToolByName(contained_by, "types_tool", None)
    if types_tool is not None:
        parentFti = types_tool.getTypeInfo(contained_by)
        if parentFti is not None:
            aliases = parentFti.getMethodAliases()
            if aliases is not None and cid in aliases.keys():
                return _("${name} is reserved.", mapping={"name": cid})

    # Lastly, we want to disallow the id of any of the tools in the portal
    # root, as well as any object that can be acquired via portal_skins.
    # However, we do want to allow overriding of *content* in the object's
    # parent path, including the portal root.

    if cid == "index_html":
        # always allow index_html
        return
    portal = getSite()
    if portal and cid in portal.contentIds():
        # Fine to use the same id as a *content* item from the root.
        return
    # It is allowed to give an object the same id as another
    # container in it's acquisition path as long as the
    # object is outside the portal.
    outsideportal = getattr(aq_parent(portal), cid, None)
    insideportal = getattr(portal, cid, None)
    if (
        insideportal is not None
        and outsideportal is not None
        and aq_base(outsideportal) == aq_base(insideportal)
    ):
        return
    # but not other things
    if getattr(portal, cid, None) is not None:
        return _("${name} is reserved.", mapping={"name": cid})


def get_user_friendly_types(types_list=None):
    """List of types which are considered "user friendly" for search and selection purposes.

    This is the list of types available in the portal, minus those defined in the
    `types_not_searched` property in registry, if it exists.

    If typesList is given, this is used as the base list;
    else all types from portal_types are used.
    """
    registry = getUtility(IRegistry)
    search_settings = registry.forInterface(ISearchSchema, prefix="plone")
    ttool = getUtility(ITypesTool)
    types = set(ttool.keys())
    if types_list:
        types = {t for t in types_list if t in types}
    friendly_types = types - set(search_settings.types_not_searched)
    return list(friendly_types)


def unrestricted_construct_instance(type_name, container, id, *args, **kw):
    """Create an object without performing security checks

    invokeFactory and fti.constructInstance perform some security checks
    before creating the object. Use this function instead if you need to
    skip these checks.

    This method uses
    CMFCore.TypesTool.FactoryTypeInformation._constructInstance
    to create the object without security checks.
    """
    id = str(id)
    typesTool = getToolByName(container, 'portal_types')
    fti = typesTool.getTypeInfo(type_name)
    if not fti:
        raise ValueError('Invalid type %s' % type_name)

    return fti._constructInstance(container, id, *args, **kw)



from zope.interface import Interface


class IMigrationTool(Interface):
    """Handles migrations between Plone releases."""

    def getInstanceVersion():
        """The version this instance of Plone is on."""

    def setInstanceVersion(version):
        """The version this instance of Plone is on."""

    def getFileSystemVersion():
        """The version the filesystem code of Plone is on."""

    def needUpgrading():
        """Need upgrading?"""

    def coreVersions():
        """Useful core version information."""

    def coreVersionsList():
        """Useful core version information."""

    def needUpdateRole():
        """Do roles need to be updated?"""

    def needRecatalog():
        """Does this thing now need recataloging?"""

    def upgrade(REQUEST=None, dry_run=None, swallow_errors=1):
        """Perform the upgrade."""



from zope.interface import Interface


class IPloneBaseTool(Interface):
    """Marker interface for plone tools"""


class IPloneTool(Interface):
    """Marker interface for the plone utils tool."""


class IPloneCatalogTool(Interface):
    """Marker interface for Plone's catalog tool"""



from zope.interface import Interface


# Language-support
class ILanguage(Interface):
    def get_language(self):
        """return the contents language"""

    def set_language(self):
        """return the contents language"""



from zope.i18nmessageid import MessageFactory


# create messages in the plone domain
PloneMessageFactory = MessageFactory("plone")

# create messages in the plonelocales domain
PloneLocalesMessageFactory = MessageFactory("plonelocales")



from .. import PloneMessageFactory as _
from zope import schema
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import zope.component
import zope.interface


OVERRIDE_RESOURCE_DIRECTORY_NAME = "resource_overrides"


class IResourceRegistry(zope.interface.Interface):

    # DEPRECATED

    url = schema.ASCIILine(title=_("Resources base URL"), required=False)

    js = schema.ASCIILine(title=_("Main js file"), required=False)

    css = schema.List(
        title=_("CSS/LESS files"),
        value_type=schema.ASCIILine(title=_("URL")),
        default=[],
        required=False,
    )

    init = schema.ASCIILine(title=_("Init instruction for shim"), required=False)

    deps = schema.ASCIILine(
        title=_("Dependencies for shim"),
        description=_("Comma separated values of resource for shim"),
        required=False,
    )

    export = schema.ASCIILine(title=_("Export vars for shim"), required=False)

    conf = schema.Text(
        title=_("Configuration in JSON for the widget"),
        description=_("Should be accessible on @@getWCconfig?id=name"),
        required=False,
    )


class IBundleRegistry(zope.interface.Interface):

    jscompilation = schema.ASCIILine(
        title=_("URL of the last js compilation"), required=False
    )

    csscompilation = schema.ASCIILine(
        title=_("URL of the last css compilation"), required=False
    )

    expression = schema.ASCIILine(
        title=_("Expression to render"),
        description=_(
            "In case its a bundle we can have a condition to render it (it "
            "does not apply if the bundle is merged)."
        ),
        required=False,
    )

    enabled = schema.Bool(title=_("It's enabled?"), default=True, required=False)

    depends = schema.ASCIILine(
        title=_("Depends on another bundle"),
        description=_(
            "In case you want to be the last: *, in case its the first should be empty"
        ),
        required=False,
    )

    load_async = schema.Bool(
        title=_("Load asynchronously"),
        description=_(
            "Load the JavaScript files asynchronously by adding an ``async`` attribute to the script tag."
        ),
        default=False,
        required=False,
    )

    load_defer = schema.Bool(
        title=_("Load deferred"),
        description=_(
            "Load the JavaScript files deferred after the document has been parsed but before ``DOMContentLoaded`` by adding a ``defer`` attribute to the script tag."
        ),
        default=False,
        required=False,
    )

    # DEPRECATED IN Plone 6, for now keep for BBB
    compile = schema.Bool(
        title=_("(DEPRECATED) Does your bundle contains any RequireJS or LESS file?"),
        description=_(
            "If its true and you modify this bundle you need to build it before production"
        ),
        default=True,
        required=False,
    )

    resources = schema.List(
        title=_("(DEPRECATED) Loaded resources"),
        description=_(
            "The resources that are going to be loaded on this bundle in order"
        ),
        value_type=schema.ASCIILine(title=_("Resource name")),
        required=False,
    )

    last_compilation = schema.Datetime(
        title=_("(DEPRECATED) Last compiled date"),
        description=_("Date time of the last compilation of this bundle"),
        required=False,
    )

    develop_javascript = schema.Bool(
        title=_("(DEPRECATED) Develop JavaScript"), default=False
    )

    develop_css = schema.Bool(title=_("(DEPRECATED) Develop CSS"), default=False)

    stub_js_modules = schema.List(
        title=_("(DEPRECATED) Stub JavaScript modules"),
        description=_(
            "Define list of modules that will be defined empty "
            "on RequireJS build steps to prevent loading modules multiple times."
        ),
        value_type=schema.ASCIILine(title=_("Resource name")),
        required=False,
        missing_value=[],
        default=[],
    )

    merge_with = schema.Choice(
        title=_("(DEPRECATED) Merge with"),
        description=_(
            "In production mode, bundles are merged together to reduce the "
            "quantity of JS and CSS resources loaded by the browser. Choose "
            "'default' if this bundle must be available for all the visitors, "
            "choose 'logged-in' if it must be available for logged-in users "
            "only, or leave it empty if it must not be merged."
        ),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("", "", _("")),
                SimpleTerm("default", "default", "default"),
                SimpleTerm("logged-in", "logged-in", "logged-in"),
            ]
        ),
        default="",
        required=False,
    )



""" CMFPlone product permissions """
from AccessControl import ModuleSecurityInfo


security = ModuleSecurityInfo("Products.CMFPlone.permissions")

security.declarePublic("AccessContentsInformation")
from Products.CMFCore.permissions import AccessContentsInformation


security.declarePublic("AddPortalContent")
from Products.CMFCore.permissions import AddPortalContent


security.declarePublic("AddPortalFolders")
from Products.CMFCore.permissions import AddPortalFolders


security.declarePublic("AddPortalMember")
from Products.CMFCore.permissions import AddPortalMember


security.declarePublic("DeleteObjects")
from Products.CMFCore.permissions import DeleteObjects


security.declarePublic("FTPAccess")
from Products.CMFCore.permissions import FTPAccess


security.declarePublic("ListFolderContents")
from Products.CMFCore.permissions import ListFolderContents


security.declarePublic("ListPortalMembers")
from Products.CMFCore.permissions import ListPortalMembers


security.declarePublic("ListUndoableChanges")
from Products.CMFCore.permissions import ListUndoableChanges


security.declarePublic("ManagePortal")
from Products.CMFCore.permissions import ManagePortal


security.declarePublic("ManageProperties")
from Products.CMFCore.permissions import ManageProperties


security.declarePublic("ManageUsers")
from Products.CMFCore.permissions import ManageUsers


security.declarePublic("ModifyPortalContent")
from Products.CMFCore.permissions import ModifyPortalContent


security.declarePublic("ReplyToItem")
from Products.CMFCore.permissions import ReplyToItem


security.declarePublic("RequestReview")
from Products.CMFCore.permissions import RequestReview


security.declarePublic("ReviewPortalContent")
from Products.CMFCore.permissions import ReviewPortalContent


security.declarePublic("SetOwnPassword")
from Products.CMFCore.permissions import SetOwnPassword


security.declarePublic("SetOwnProperties")
from Products.CMFCore.permissions import SetOwnProperties


security.declarePublic("UndoChanges")
from Products.CMFCore.permissions import UndoChanges


security.declarePublic("View")
from Products.CMFCore.permissions import View


security.declarePublic("ViewManagementScreens")
from Products.CMFCore.permissions import ViewManagementScreens



"""
Collection of i18n and l10n utility methods.
"""
from Acquisition import aq_acquire
from DateTime import DateTime
from DateTime.interfaces import IDateTime
from plone.registry.interfaces import IRegistry
from zope.component import getUtility
from zope.i18n import translate
from zope.i18n.locales import locales
from zope.publisher.interfaces.browser import IBrowserRequest

import logging
import re


logger = logging.getLogger(__name__)

datetime_formatvariables = {"H", "I", "m", "d", "M", "p", "S", "Y", "y", "Z"}
name_formatvariables = {"a", "A", "b", "B"}
all_formatvariables = datetime_formatvariables | name_formatvariables
_all_regexp_set = ",".join(all_formatvariables)
# regexp to split up ${X} format strings
_interp_regex = re.compile(
    r"(?<!\$)(\$(?:[%(n)s]|{[%(n)s]}))" % ({"n": _all_regexp_set})
)
# regexp to detect if this is a strftime format string
_dt_format_string_regexp = re.compile(rf"\%([{_all_regexp_set}])")

# those are from DateTime.DateTime, but we must not rely on its internal
# structures, so here a copy:
ENGLISH_NAMES = {
    "_days": (
        "Sunday",
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
    ),
    "_days_a": ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"),
    "_days_p": ("Sun.", "Mon.", "Tue.", "Wed.", "Thu.", "Fri.", "Sat."),
    "_months": (
        "",
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ),
    "_months_a": (
        "",
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ),
    "_months_p": (
        "",
        "Jan.",
        "Feb.",
        "Mar.",
        "Apr.",
        "May",
        "June",
        "July",
        "Aug.",
        "Sep.",
        "Oct.",
        "Nov.",
        "Dec.",
    ),
}

# The following are helper methods to change the default date and time formats
# for a specific locale. These locale dependent formats are used in the
# date/time widgets to determine the format and decide if a 24 hour or 12 hour
# AM/PM widget should be used. If 'a' is part of the pattern the AM/PM widget
# will be used, otherwise a 24 hour clock.
#
# localeid is a tuple of the form: (language, country, variant)
# for example: (None, ) or ('en', ) or ('en', 'US', None)
#
# value is in the format described by zope.i18n.interfaces.IDateTimeFormat
# for example u'yyyy-MM-dd' or u'HH:mm:ss'
#
# Note that this is a different format than used for the other methods in
# this module.
#
# locales uses a module level cache, so any changes you make with these
# methods will apply to the entire process and only need to be made once.
# You can use them in any code imported at startup, for example in a packages
# __init__ method.
#
# In order to use a 24 hour clock for English speakers, you would do:
#
# from Products.CMFPlone import i18nl10n
# i18nl10n.setDefaultTimeFormat(('en',), u'HH:mm:ss')


def setDefaultDateFormat(localeid, value):
    gregorian = locales.getLocale(*localeid).dates.calendars["gregorian"]
    date_format = gregorian.dateFormats["medium"].formats[None]
    date_format.pattern = value


def setDefaultTimeFormat(localeid, value):
    gregorian = locales.getLocale(*localeid).dates.calendars["gregorian"]
    time_format = gregorian.timeFormats["medium"].formats[None]
    time_format.pattern = value


def utranslate(
    domain, msgid, mapping=None, context=None, target_language=None, default=None
):
    # We used to pass an object as context.
    if not IBrowserRequest.providedBy(context):
        context = aq_acquire(context, "REQUEST")
    # The signature of zope.i18n's translate has the msgid and domain switched
    return translate(
        msgid,
        domain=domain,
        mapping=mapping,
        context=context,
        target_language=target_language,
        default=default,
    )


def get_formatstring_from_registry(msgid):
    """If the Enabled record is True, return a format string."""
    registry = getUtility(IRegistry)
    name_root = "Products.CMFPlone.i18nl10n.override_dateformat."
    if registry.get(name_root + "Enabled", False) is False:
        return None
    # msgid: "date_format_long", "date_format_short", or "time_format"
    record_name = name_root + msgid
    return registry.get(record_name, None)


def ulocalized_time(
    time,
    long_format=None,
    time_only=False,
    context=None,
    domain="plonelocales",
    request=None,
    target_language=None,
):
    """unicode aware localized time method (l10n)"""

    if time_only:
        msgid = "time_format"
    elif long_format:
        msgid = "date_format_long"
    else:
        msgid = "date_format_short"

    # NOTE: this requires the presence of three msgids inside the translation
    #       catalog date_format_long, date_format_short, and time_format
    #       These msgids are translated using interpolation.
    #       The variables used here are the same as used in the strftime
    #       formating.
    #       Supported are:
    #           %A, %a, %B, %b, %H, %I, %m, %d, %M, %p, %S, %Y, %y, %Z
    #       Each used as variable in the msgstr without the %.
    #       For example: "${A} ${d}. ${B} ${Y}, ${H}:${M} ${Z}"
    #       Each language dependend part is translated itself as well.

    # From http://docs.python.org/lib/module-time.html
    #
    # %a    Locale's abbreviated weekday name.
    # %A        Locale's full weekday name.
    # %b        Locale's abbreviated month name.
    # %B        Locale's full month name.
    # %d        Day of the month as a decimal number [01,31].
    # %H        Hour (24-hour clock) as a decimal number [00,23].
    # %I        Hour (12-hour clock) as a decimal number [01,12].
    # %m        Month as a decimal number [01,12].
    # %M        Minute as a decimal number [00,59].
    # %p        Locale's equivalent of either AM or PM.
    # %S        Second as a decimal number [00,61].
    # %y        Year without century as a decimal number [00,99].
    # %Y        Year with century as a decimal number.
    # %Z        Time zone name (no characters if no time zone exists).

    mapping = {}
    # convert to DateTime instances. Either a date string or
    # a DateTime instance needs to be passed.
    if not IDateTime.providedBy(time):
        try:
            time = DateTime(time)
        except:
            logger.debug(f"Failed to convert {time} to a DateTime object")
            return None

    if context is None:
        # when without context, we cannot do very much.
        return time.ISO8601()

    if request is None:
        request = aq_acquire(context, "REQUEST")

    if long_format and isinstance(long_format, str):
        # 0. If a format is explicitly passed in the long_format argument, this wins.
        #    Hopefully something like this: "${a} ${d} ${b} ${Y}""
        formatstring = long_format
    else:
        # 1. if our Enabled flag in the configuration registry is set,
        # the format string there should override the translation machinery
        formatstring = get_formatstring_from_registry(msgid)

    if formatstring is not None:
        if _dt_format_string_regexp.findall(formatstring):
            # This is classic strftime formatting with percentages instead of dollars,
            # e.g. %Y instead of ${Y} for the year.  This means we cannot do further
            # i18n/l10n for translating week days or month names.  Python will do
            # translation using the current locale.
            return time.strftime(formatstring)
    else:
        # 2. the normal case: translation machinery,
        # that is the ".../LC_MESSAGES/plonelocales.po" files
        formatstring = translate(
            msgid, domain, mapping, request, target_language=target_language
        )

    # 3. if both failed, fall back to hardcoded ISO style
    if formatstring == msgid:
        if msgid == "date_format_long":
            formatstring = "%Y-%m-%d %H:%M"  # 2038-01-19 03:14
        elif msgid == "date_format_short":
            formatstring = "%Y-%m-%d"  # 2038-01-19
        elif msgid == "time_format":
            formatstring = "%H:%M"  # 03:14
        else:
            formatstring = "[INTERNAL ERROR]"
        return time.strftime(formatstring)

    # get the format elements used in the formatstring
    formatelements = {el[2:-1] for el in _interp_regex.findall(formatstring)}

    # add used elements to mapping
    elements = formatelements & datetime_formatvariables
    for key in elements:
        mapping[key] = time.strftime("%" + key)

    # add weekday name, abbr. weekday name, month name, abbr month name
    name_elements = formatelements & name_formatvariables
    if {"a", "A"} & name_elements:
        weekday = int(time.strftime("%w"))  # weekday, sunday = 0
        if "a" in name_elements:
            mapping["a"] = weekdayname_msgid_abbr(weekday)
        if "A" in name_elements:
            mapping["A"] = weekdayname_msgid(weekday)
    if {"b", "B"} & name_elements:
        monthday = int(time.strftime("%m"))  # month, january = 1
        if "b" in name_elements:
            mapping["b"] = monthname_msgid_abbr(monthday)
        if "B" in name_elements:
            mapping["B"] = monthname_msgid(monthday)

    # translate translateable elements
    for key in name_elements:
        mapping[key] = translate(
            mapping[key],
            domain,
            context=request,
            default=mapping[key],
            target_language=target_language,
        )

    # translate the time string
    return translate(
        formatstring, domain, mapping, request, target_language=target_language
    )


def _numbertoenglishname(number, format=None, attr="_days"):
    # returns the english name of day or month number
    # starting with Sunday == 0
    # and January = 1
    # format is either None, 'a' or 'p')
    #   None  means full name (January, February, ...)
    #   'a' means abbreviated (Jan, Feb, ..)
    #   'p' means abbreviated with . (dot) at end (Jan., Feb., ...)

    number = int(number)
    if format is not None:
        attr = f"{attr}_{format}"
    return ENGLISH_NAMES[attr][number]


def monthname_english(number, format=None):
    # returns the english name of month with number
    return _numbertoenglishname(number, format=format, attr="_months")


def weekdayname_english(number, format=None):
    # returns the english name of week with number
    return _numbertoenglishname(number, format=format, attr="_days")


def monthname_msgid(number):
    # returns the msgid for monthname
    # use to translate to full monthname (January, February, ...)
    # e.g. month_jan, month_feb, ...
    return "month_%s" % monthname_english(number, format="a").lower()


def monthname_msgid_abbr(number):
    # returns the msgid for the abbreviated monthname
    # use to translate to abbreviated format (Jan, Feb, ...)
    # e.g. month_jan_abbr, month_feb_abbr, ...
    return "month_%s_abbr" % monthname_english(number, format="a").lower()


def weekdayname_msgid(number):
    # returns the msgid for the weekdayname
    # use to translate to full weekdayname (Monday, Tuesday, ...)
    # e.g. weekday_mon, weekday_tue, ...
    return "weekday_%s" % weekdayname_english(number, format="a").lower()


def weekdayname_msgid_abbr(number):
    # returns the msgid for abbreviated weekdayname
    # use to translate to abbreviated format (Mon, Tue, ...)
    # e.g. weekday_mon_abbr, weekday_tue_abbr, ...
    return "weekday_%s_abbr" % weekdayname_english(number, format="a").lower()


def weekdayname_msgid_short(number):
    # return the msgid for short weekdayname
    # use to translate to 2 char format (Mo, Tu, ...)
    # e.g. weekday_mon_short, weekday_tue_short, ...
    return "weekday_%s_short" % weekdayname_english(number, format="a").lower()



from zope.interface import Interface


class INonInstallable(Interface):
    def getNonInstallableProfiles():
        """Returns a list of profiles that should not be available for
        installation at portal creation time.

        The usual use-case is to prevent extension profiles from showing up,
        that will be installed as part of the site creation anyways.
        """

    def getNonInstallableProducts():
        """Returns a list of products that should not be available for
        installation.
        """
